import streamlit as st

from ovo import (
    local_scheduler,
    storage,
    config,
)
from ovo.core.database.models_rfdiffusion import (
    RFdiffusionScaffoldDesignWorkflow,
    MODEL_WEIGHTS_SCAFFOLD,
)
from ovo.app.components import molstar_custom_component, StructureVisualization
from ovo.app.components.history_components import history_dropdown_component
from ovo.app.components.input_components import pdb_input_component, sequence_selection_fragment, initialize_workflow
from ovo.app.components.molstar_custom_component import ContigsParser
from ovo.app.components.navigation import show_prev_next_sections
from ovo.app.components.preview_components import visualize_rfdiffusion_preview, contigs_organizer_fragment
from ovo.app.components.scheduler_components import wait_with_statusbar
from ovo.app.components.submission_components import (
    pool_submission_inputs,
    show_rfdiffusion_advanced_settings,
    review_workflow_submission,
)
from ovo.app.pages import jobs_page, designs_page
from ovo.app.utils.page_init import initialize_page
from ovo.core.auth import get_username
from ovo.core.database import descriptors_refolding
from ovo.core.logic.design_logic_rfdiffusion import submit_rfdiffusion_preview
from ovo.core.utils.formatting import get_hashed_path_for_bytes


@st.fragment
def intro_step():
    # Initialize the workflow object in session state
    initialize_workflow(__file__, RFdiffusionScaffoldDesignWorkflow.name)

    with st.container(width=850):
        st.markdown(
            f"""
This workflow enables designing a new protein structure that supports fixed segments (*motifs*) from an existing structure. 
It implements and extends the end-to-end workflow for motif scaffolding from RFdiffusion<sup>[1]</sup>.

Possible use cases include:

- Scaffolding a new enzyme around a defined active-site geometry
- Truncating a region of a protein formed from multiple discontinuous regions of the sequence
- Designing a new scaffold that stabilizes that fixed region and enhances solubility (e.g. by removing an unwanted hydrophobic interface)
            
This workflow combines RFdiffusion<sup>[1]</sup> for de novo backbone design, LigandMPNN<sup>[2]</sup> for sequence design 
and side-chain rotamer prediction and AlphaFold2<sup>[3]</sup> or ESMFold<sup>[4]</sup> for independent validation.

- **Input**
    - Protein structure (PDB file or UniProt ID)
- **Main parameters**
    - "Contig" string describing stretches of residues to keep and stretches of residues to generate
    - (Optional) Stretches of residues to keep in the structure but redesign in sequence (inpainting)
- **Output**
    - Designs — protein sequences and structures
    - Design descriptors
        - Refolding metrics
        - Backbone metrics
        - Sequence composition

### Features

After inputting your structure, you'll first select segments of "fixed" residues to keep in the design.
Next, you will defined *contig* that specifies how the fixed segments are connected and how many residues are designed in between 
(and in C-/N-termini). You can generate a quick RFdiffusion preview to verify your contig.
You can choose to redesign some of the fixed residues in sequence while keeping their backbone coordinates fixed (inpainting).
Finally, you will set the number of designs to generate and optional advanced settings and submit the workflow.

Once the workflow has finished, all designs are copied from the workflow working directory to the OVO *storage* and registered in the OVO *database*.
You will be able to view all designs and their descriptors in the {jobs_page.title} results page and adjust thresholds to select *accepted designs*.
The accepted designs (usually only around 0.1% to 1% of all generated designs in typical RFdiffusion workflows) can be further 
analyzed in the {designs_page.title} page.

##### Backbone metrics

- Computed for the RFdiffusion backbone design
- Secondary structure composition
    - Percentages of helix, sheet, loop (the remaining residues); computed using PyDSSP<sup>[5]</sup>
- Radius of gyration
    - A measure of compactness of the protein structure, defined as the root mean square distance of the atoms from the center of mass (using Cα atoms only).

##### Sequence composition

- Percentage of individual amino acids
- Charge at pH 7.4 and 5.5, isoelectric point
- Sequence entropy, highlighting sequences with compositionally biased regions
- Aromaticity
- And more; computed using BioPython ProteinAnalysis<sup>[6]</sup>

##### Refolding metrics

- **AlphaFold2 initial guess<sup>[3][7][8]</sup>**
    - **{descriptors_refolding.AF2_PRIMARY_PAE.name}** — Predicted Aligned Error <sup>[9]</sup> averaged over the residue–residue matrix, the lower the more confident the model is about overall design. <sup>*</sup>
    - **{descriptors_refolding.AF2_PRIMARY_DESIGN_RMSD.name}** — Cα RMSD between RFdiffusion output and AF2 prediction, measuring the agreement between the design method and the structure prediction method.
    - **{descriptors_refolding.AF2_PRIMARY_NATIVE_MOTIF_RMSD.name}** — All-atom RMSD between motif in your input structure and the motif as predicted by AF2, ensuring the original motif was neither disrupted by RFdiffusion or by the AlphaFold2 prediction.
- **ESMFold (if enabled)**
    - **{descriptors_refolding.ESMFOLD_PAE.name}** — as in AF2 PAE
    - **{descriptors_refolding.ESMFOLD_DESIGN_BACKBONE_RMSD.name}** - Cα RMSD between whole RFdiffusion output and ESMFold prediction
    - **{descriptors_refolding.ESMFOLD_DESIGN_ALL_ATOM_RMSD.name}** - All-atom RMSD between whole RFdiffusion output and ESMFold prediction
    - Note that "Native Motif RMSD" is not computed for ESMFold yet

<sup>*</sup> Note that should your design not have a static structure (e.g. domains whose relative positions can change), PAE will be higher.

In RFdiffusion<sup>[1]</sup> in silico success for scaffold design was defined as 
AF2&nbsp;PAE&nbsp; <&nbsp;5.0, 
AF2&nbsp;Design&nbsp;RMSD <&nbsp;2.0, 
and AF2&nbsp;Native&nbsp;Motif RMSD <&nbsp;1.0.
         
For both AlphaFold2 and ESMFold we also include pLDDT and pTM scores.

### Limitations

- ⚠️ **RFdiffusion workflows are computationally intensive** and typically require many designs (10,000 backbones or more) to obtain a few high-quality designs. 
     Such a workflow can take weeks of GPU time (finishing under a day thanks to parallelization). 
     Please consider starting with a small number of designs (100-200 backbones) to verify your inputs and settings before scaling up.
- ❌ **Cofactors are not supported**; only protein entities
- ❌ **SolubleMPNN weights are currently not supported**
- ⚠️ **Design of multi-chain complexes is supported but experimental**, some UI components may not behave as expected
- ⚠️ **This page enables running the end-to-end workflow**. Individual steps like ProteinMPNN redesign or AlphaFold2 scoring 
    alone cannot currently be executed through the user interface.

### Methods
                    
RFdiffusion generates a backbone → LigandMPNN generates sequences and side-chain rotamers → AlphaFold2 and/or ESMFold evaluates the design by predicting structures 
from the designed sequence comparing them to the RFdiffusion backbone (refolding, sometimes called self-consistency or designability test).

|             | Purpose     |  Input | Output |
|-------------|-------------|-------------|-------------|
| RFdiffusion | Backbone generation       | Residue stretches to fix (structure and, by default, sequence) and stretches to design | Backbone structure |
| LigandMPNN  | Sequence generation and side-chain rotamer prediction | Backbone structure | Sequence and structure with side-chains |
| AlphaFold2  | Refolding evaluation (via structure prediction)      | Designed sequence, structure (from LigandMPNN, see below) | Structure |
| ESMFold     | Refolding evaluation (via structure prediction)      | Designed sequence only | Structure |

**Details**
                
RFdiffusion generation
- RFdiffusion does not explicitly predict sequences for designed regions; however it uses an implicit notion of sequence while generating backbones. The authors note it could have been trained to co-design sequence and backbone, but they found the RFdiffusion + ProteinMPNN workflow works well. Using ProteinMPNN also enables generating multiple sequences for the same backbone.
- With `inpaint_seq` parameter, RFdiffusion holds only the backbone coordinates fixed while allowing the sequence to change by masking it out for the specified stretches (e.g. you can keep the backbone of a whole beta sheet and redesign the residues on one side).
- Note that RFdiffusion discards all side-chain atoms in the output PDB - including regions of fixed segments and chains. These are reconstructed by LigandMPNN and by the AlphaFold2 prediction.

LigandMPNN (vs. ProteinMPNN<sup>[10]</sup>)
- Incorporates explicit atomic context for small molecules and metal ions, improving sequence prediction when ligands are present. (However ligands are not currently supported.)
- Predicts side-chain rotamers and can fix side-chain conformations for selected residues; ProteinMPNN returns sequence only.
- Similar training set

AlphaFold2 protocol
- Initial guess: the design structure (from LigandMPNN) is supplied to the structure module before the first recycle, influencing the starting conformation. This is always enabled.
- Template input: the design structure is provided to every recycle, giving a stronger persistent structural context.
- Runs in single-sequence mode (no MSA), achieving speeds of around 10 seconds for a 100aa design on a low-end T4 GPU. 
  However, scales quadratically with sequence length (same as regular AlphaFold2).

### References
1. [RFdiffusion](https://www.nature.com/articles/s41586-023-06415-8), (and [Motif Scaffolding GitHub guide](https://github.com/RosettaCommons/RFdiffusion?tab=readme-ov-file#motif-scaffolding))
2. [LigandMPNN](https://www.nature.com/articles/s41592-025-02626-1)
3. [AlphaFold2](https://www.nature.com/articles/s41586-021-03819-2)
4. [ESMFold](https://www.science.org/doi/10.1126/science.ade2574)
5. [PyDSSP](https://github.com/ShintaroMinami/PyDSSP)
6. [BioPython ProteinAnalysis](https://biopython.org/docs/latest/api/Bio.SeqUtils.ProtParam.html#Bio.SeqUtils.ProtParam.ProteinAnalysis)
7. [AlphaFold2 initial guess](https://www.nature.com/articles/s41467-023-38328-5)
8. [ColabDesign](https://github.com/sokrypton/ColabDesign)
9. [Predicted Aligned Error (PAE)](https://www.ebi.ac.uk/training/online/courses/alphafold/inputs-and-outputs/evaluating-alphafolds-predicted-structures-using-confidence-scores/pae-a-measure-of-global-confidence-in-alphafold-predictions/)
10. [ProteinMPNN](https://www.science.org/doi/10.1126/science.add2187)
""",
            unsafe_allow_html=True,
        )


@st.fragment
def input_step():
    workflow: RFdiffusionScaffoldDesignWorkflow = st.session_state.workflows[__file__]

    left, _, right = st.columns([4, 1, 3])
    with left:
        st.subheader("Input structure")
        if new_pdb_input := pdb_input_component(workflow.input_name):
            input_name, pdb_input_bytes = new_pdb_input
            file_hash = get_hashed_path_for_bytes(pdb_input_bytes)

            workflow.rfdiffusion_params.input_pdb = storage.store_file_str(
                pdb_input_bytes.decode(),
                f"project/{st.session_state.project.id}/inputs/{file_hash}/{input_name}.pdb",
                overwrite=False,
            )
            workflow.input_name = input_name
            workflow.selected_segments = []
            # TODO use presets to initialize this instead
            workflow.refolding_params.primary_test = "af2_model_1_ptm_ft_3rec"
    with right:
        history_dropdown_component(__file__, workflow_name=RFdiffusionScaffoldDesignWorkflow.name)

    if not workflow.rfdiffusion_params.input_pdb:
        # No input structure provided yet, stop here
        return

    st.subheader(workflow.input_name)

    molstar_custom_component(
        structures=[
            StructureVisualization(pdb=storage.read_file_str(workflow.rfdiffusion_params.input_pdb), color="chain-id")
        ],
        key="input_structure",
        width=700,
        height=400,
    )


@st.fragment()
def selection_step():
    workflow: RFdiffusionScaffoldDesignWorkflow = st.session_state.workflows[__file__]

    st.subheader("Select residues to keep")

    if not workflow.rfdiffusion_params.input_pdb:
        st.error("Please provide an input structure in the input structure step.")
        return

    sequence_selection_fragment(__file__, workflow.input_name)


@st.fragment()
def contig_preview_step():
    workflow: RFdiffusionScaffoldDesignWorkflow = st.session_state.workflows[__file__]
    st.subheader("Enter contig and generate preview")

    st.write(
        """
        The contig defines which parts of the structure will be preserved,
        in which order they will be connected, and how many residues will be designed in between.
        """
    )

    if not workflow.rfdiffusion_params.input_pdb:
        st.error("Please provide an input structure in the input structure step.")
        return

    # Input parameters for the preview
    contigs_organizer_fragment(__file__, pdb_input_string=storage.read_file_str(workflow.rfdiffusion_params.input_pdb))

    # Generate preview
    st.write("#### Generate preview")

    help_column = st.columns([2, 1])[0]

    with st.columns(3)[0]:
        if "preview_timesteps" not in st.session_state:
            st.session_state.preview_timesteps = 5
        new_timesteps = st.slider(
            "Num RFdiffusion timesteps (T)",
            min_value=1,
            max_value=20,
            value=st.session_state.preview_timesteps,
            key="timesteps_input",
        )
        if new_timesteps and new_timesteps != st.session_state.preview_timesteps:
            st.session_state.preview_timesteps = new_timesteps
            # Clear previous preview if settings changed
            workflow.preview_job_id = None

    with help_column:
        st.write(f"""
        Generate a quick RFdiffusion preview of the design with reduced number of timesteps
        ({st.session_state.preview_timesteps}/50) to verify your inputs. This step is optional.

        This should take from 30 seconds to a few minutes depending on the length of the protein.
        """)

    if st.button(":material/wand_stars: Generate preview"):
        with st.spinner("Submitting RFdiffusion job..."):
            workflow.preview_job_id = submit_rfdiffusion_preview(workflow, timesteps=st.session_state.preview_timesteps)

    # Check if needed parameters are set
    if not workflow.preview_job_id:
        return

    if workflow.rfdiffusion_params.contig is None:
        return

    # Wait until task is done
    result = wait_with_statusbar(local_scheduler, workflow.preview_job_id, label="Running RFdiffusion...")

    # Exit if job failed
    if not result:
        return

    visualize_rfdiffusion_preview(workflow, output_dir=local_scheduler.get_output_dir(workflow.preview_job_id))


@st.fragment()
def inpainting_step():
    workflow: RFdiffusionScaffoldDesignWorkflow = st.session_state.workflows[__file__]

    st.subheader("Sequence inpainting")

    st.write(
        "Select input structure residues that will be kept in the structure but redesigned with ProteinMPNN. This step is optional."
    )

    if not workflow.rfdiffusion_params.input_pdb:
        st.error("Please provide an input structure in the input structure step.")
        return

    if not workflow.rfdiffusion_params.contig:
        st.error("Please provide a contig in the previous step.")
        return

    parsed_contig = ContigsParser().parse_contigs_str(workflow.rfdiffusion_params.contig)
    fixed_segments = [seg for seg in parsed_contig if seg.type == "fixed"]

    sequence_selection_fragment(__file__, workflow.input_name, fixed_segments=fixed_segments, inpainting=True)


@st.fragment()
def settings_step():
    st.subheader("Settings")

    workflow: RFdiffusionScaffoldDesignWorkflow = st.session_state.workflows[__file__]

    if not workflow.rfdiffusion_params.input_pdb:
        st.error("Please provide an input structure in the input structure step.")
        return

    if not workflow.rfdiffusion_params.contig:
        st.error("Contig has not been computed. Please fill in all inputs in previous step.")
        return

    pool_submission_inputs(__file__)

    contig = st.text_input(
        "Contig",
        placeholder="A123-456/10/A567-890/...",
        value=workflow.rfdiffusion_params.contig,
        key="input_contig",
    )

    if check_contig_parsed(contig):
        workflow.rfdiffusion_params.contig = contig

    with st.columns([1, 2])[0]:
        workflow.rfdiffusion_params.model_weights = st.selectbox(
            "Model weights",
            help="Use 'active site' model weights to hold better selected residues specified in the contig.",
            index=MODEL_WEIGHTS_SCAFFOLD.index(workflow.rfdiffusion_params.model_weights)
            if workflow.rfdiffusion_params.model_weights
            else 0,
            key="active_site",
            options=MODEL_WEIGHTS_SCAFFOLD,
        )

        is_admin = get_username() in config.auth.admin_users
        workflow.rfdiffusion_params.num_designs = st.number_input(
            "Number of structure designs (RFdiffusion backbones)",
            min_value=1,
            max_value=config.props.rfdiffusion_backbones_limit_admin
            if is_admin
            else config.props.rfdiffusion_backbones_limit,
            value=workflow.rfdiffusion_params.num_designs,
            key="num_designs",
        )

        workflow.protein_mpnn_params.num_sequences = st.number_input(
            "Number of sequence designs per backbone (ProteinMPNN)",
            min_value=1,
            max_value=config.props.mpnn_sequences_limit,
            value=workflow.protein_mpnn_params.num_sequences,
            key="num_sequences",
        )

    show_rfdiffusion_advanced_settings(workflow)


@st.fragment
def review_step():
    st.subheader("Review settings")

    review_workflow_submission(__file__)


def check_contig_parsed(contig: str | None, verbose: bool = False) -> bool:
    parser = ContigsParser()
    if not len(contig.split(" ")) == 1:
        if verbose:
            st.warning("Support of multi-chain scaffold design is experimental, proceed with caution")

    if contig.islower():
        st.warning("Input contigs are lowercase. Workflow can behave unexpectedly.")

    try:
        _ = parser.parse_contigs_str(contig) if contig else None
        return True
    except Exception as e:
        if verbose:
            st.error(
                f"There is an error in the contig specification. Please make sure you provide valid contigs. Error: {e}"
            )
        return False


if __name__ == "__main__":
    initialize_page(page_title="Scaffold design")

    show_prev_next_sections(
        key=__file__,
        title="🏗️ Scaffold design",
        sections={
            "intro": intro_step,
            "input": input_step,
            "selection": selection_step,
            "contig": contig_preview_step,
            "sequence inpainting": inpainting_step,
            "settings": settings_step,
            "review": review_step,
        },
    )



import pandas as pd
import streamlit as st
from sqlalchemy.orm.attributes import flag_modified

from ovo.app.components.custom_elements import wrapped_columns
from ovo.app.utils.cached_db import get_cached_descriptor_values, Descriptor

from plotly import express as px

from ovo import db, Threshold
from ovo.core.database import Pool, DesignJob
from ovo.core.database.descriptors import ALL_DESCRIPTORS_BY_KEY
from ovo.core.logic.design_logic import update_accepted_design_ids
from ovo.core.logic.filtering_logic import filter_designs_by_thresholds


def thresholds_and_histograms_component(
    selected_thresholds: dict[str, Threshold],
    all_design_ids: list[str],
    max_items_row=3,
) -> dict[str, Threshold]:
    """Adjust thresholds using sliders and show histograms for each descriptor, return new thresholds."""

    new_thresholds = {}
    descriptor_keys = list(selected_thresholds.keys())
    columns = wrapped_columns(len(descriptor_keys), wrap=max_items_row, divider=True, gap="large")
    for descriptor_key, column in zip(descriptor_keys, columns):
        with column:
            descriptor = ALL_DESCRIPTORS_BY_KEY[descriptor_key]

            plot_container = st.container()

            st.markdown(f"##### {descriptor.name}")
            if descriptor.description:
                st.caption(descriptor.description)

            descriptor_values = get_cached_descriptor_values(descriptor.key, design_ids=all_design_ids)

            new_thresholds[descriptor.key] = single_threshold_input_component(
                threshold=selected_thresholds[descriptor.key],
                descriptor=descriptor,
                descriptor_values=descriptor_values,
            )

            with plot_container:
                descriptor_histogram_component(
                    descriptor_values=descriptor_values,
                    descriptor=descriptor,
                    threshold=new_thresholds[descriptor.key],
                )

            if new_thresholds[descriptor.key].enabled and (num_missing := descriptor_values.isna().sum()):
                st.warning(
                    f"{num_missing}/{len(all_design_ids)} designs are missing values for this descriptor. "
                    f"These designs will be marked as NOT accepted."
                )

    return new_thresholds


def thresholds_input_component(
    selected_thresholds: dict[str, Threshold],
    max_items_row: int = 3,
) -> dict[str, Threshold]:
    """Adjust thresholds using sliders, return new thresholds."""

    new_thresholds = {}

    st.markdown("#### Acceptance thresholds")
    st.markdown(
        'Designs passing *all* the thresholds will be marked as "Accepted" '
        "and become available in the Design explorer. "
        "You can change these thresholds later."
    )

    descriptor_keys = list(selected_thresholds.keys())
    columns = wrapped_columns(len(descriptor_keys), wrap=max_items_row, divider=True)
    for descriptor_key, column in zip(descriptor_keys, columns):
        with column:
            descriptor = ALL_DESCRIPTORS_BY_KEY[descriptor_key]
            st.markdown(f"##### {descriptor.name}")
            if descriptor.description:
                st.caption(descriptor.description)
            new_thresholds[descriptor.key] = single_threshold_input_component(
                threshold=selected_thresholds[descriptor_key], descriptor=descriptor, descriptor_values=None
            )

    return new_thresholds


def single_threshold_input_component(
    threshold: Threshold, descriptor: Descriptor, descriptor_values: pd.Series | None = None
) -> Threshold:
    if not threshold.enabled:
        if st.toggle("Disabled", key=f"{descriptor.key}_toggle"):
            st.session_state.thresholds_expanded = True
            st.success("Enabling threshold...")
            min_value = descriptor.min_value  # value or None
            max_value = descriptor.max_value  # value or None
            if descriptor_values is not None and len(descriptor_values):
                min_value, max_value = descriptor_values.min(), descriptor_values.max()
            # if threshold already set, use the existing values, otherwise use the descriptor min/max
            if descriptor.min_value is not None or descriptor.max_value is not None:
                return threshold.copy(enabled=True)
            elif descriptor.comparison == "higher_is_better":
                return Threshold(min_value=min_value)
            elif descriptor.comparison == "lower_is_better":
                return Threshold(max_value=max_value)
            else:
                return Threshold(min_value=min_value, max_value=max_value)
    else:
        if not st.toggle("Enabled", value=True, key=f"{descriptor.key}_toggle"):
            st.session_state.thresholds_expanded = True
            return threshold.copy(enabled=False)

    kwargs = dict(
        min_value=float(descriptor.min_value) if descriptor.min_value is not None else None,
        max_value=float(descriptor.max_value) if descriptor.max_value is not None else None,
        disabled=not threshold.enabled,
        step=0.1,
    )
    min_value, max_value = threshold.min_value, threshold.max_value
    if descriptor.comparison == "higher_is_better":
        min_value = st.number_input(
            "Min value",
            value=float(min_value) if min_value is not None else None,
            key=f"{descriptor.key}_min",
            **kwargs,
        )
    elif descriptor.comparison == "lower_is_better":
        max_value = st.number_input(
            "Max value",
            value=float(max_value) if max_value is not None else None,
            key=f"{descriptor.key}_max",
            **kwargs,
        )
    else:
        left, right = st.columns(2)
        min_value = left.number_input(
            "Min value",
            value=float(min_value) if min_value is not None else None,
            key=f"{descriptor.key}_min",
            **kwargs,
        )
        max_value = right.number_input(
            "Max value",
            value=float(max_value) if max_value is not None else None,
            key=f"{descriptor.key}_max",
            **kwargs,
        )

    return threshold.copy(min_value=min_value, max_value=max_value)


def descriptor_histogram_component(descriptor_values: pd.Series, descriptor: Descriptor, threshold: Threshold):
    fig = px.histogram(
        x=descriptor_values,
        labels={"x": descriptor.name},
        nbins=50,
        height=300,
        width=500,
        color_discrete_sequence=["#1f77b4"] if threshold.enabled else ["#888888"],
    )
    fig.update_layout(
        yaxis_title_text="Number of designs",
    )

    if bounds := threshold.get_bounds(descriptor, descriptor_values):
        # plot accepted range in green
        fig.add_shape(
            type="rect",
            x0=bounds[0],
            x1=bounds[1],
            y0=0,
            y1=1,
            yref="paper",  # treat y0 and y1 as a fraction of the y-axis
            fillcolor="#00ff00" if threshold.enabled else "#aaaaaa",
            opacity=0.15,
            line_width=0,
        )

    # Disable zoom controls
    fig.layout.xaxis.fixedrange = True
    fig.layout.yaxis.fixedrange = True
    fig.layout.margin["b"] = 0
    fig.layout.margin["t"] = 0

    st.plotly_chart(fig, width="content")


@st.cache_data(ttl="1h")
def filter_designs_by_thresholds_cached(
    all_design_ids: list[str], thresholds: dict[str, Threshold]
) -> tuple[list[str], dict[str, int]]:
    return filter_designs_by_thresholds(
        all_design_ids=all_design_ids,
        thresholds=thresholds,
        values={
            descriptor_key: get_cached_descriptor_values(descriptor_key, design_ids=all_design_ids).to_dict()
            for descriptor_key, threshold in thresholds.items()
            if threshold.enabled
        },
    )


@st.dialog("Accept designs")
def accept_designs_dialog(
    pools: list[Pool],
    jobs: list[DesignJob],
    all_design_ids: list[str],
    new_accepted_design_ids: list[str],
    num_accepted_by_descriptor: dict[str, int],
    selected_thresholds: dict[str, Threshold],
):
    st.write(
        f"Accept **{len(new_accepted_design_ids):,} / {len(all_design_ids):,}** ({len(new_accepted_design_ids) / len(all_design_ids):.0%}) designs based on new thresholds:"
    )

    if len(pools) > 1:
        st.write(f"This will update {len(pools):,} pools: **{', '.join(p.id for p in pools)}**")

    display_current_thresholds(
        selected_thresholds=selected_thresholds,
        all_design_ids=all_design_ids,
        num_accepted_by_descriptor=num_accepted_by_descriptor,
    )

    jobs_by_id = {j.id: j for j in jobs}

    with st.columns(2)[1]:
        if st.button("Save", type="primary", width="stretch"):
            updated_pool_ids = []
            for pool in pools:
                if not pool.design_job_id:
                    continue
                job = jobs_by_id[pool.design_job_id]
                job.workflow.acceptance_thresholds = selected_thresholds
                # tell sqlalchemy that we modified the workflow object - with dataclasses this is not auto-detected
                flag_modified(job, "workflow")
                updated_pool_ids.append(pool.id)
            db.save_all(pools)
            db.save_all(jobs_by_id.values())
            update_accepted_design_ids(pool_ids=updated_pool_ids, accepted_design_ids=new_accepted_design_ids)
            st.session_state.flash_success = "Accepted designs saved successfully"
            st.rerun()


def display_current_thresholds(
    selected_thresholds: dict[str, Threshold], all_design_ids: list[str], num_accepted_by_descriptor: dict[str, int]
):
    threshold_labels = []
    for descriptor_key, threshold in selected_thresholds.items():
        descriptor = ALL_DESCRIPTORS_BY_KEY[descriptor_key]
        if not threshold.enabled:
            continue
        num_accepted = num_accepted_by_descriptor[descriptor_key]
        fraction_accepted = num_accepted / len(all_design_ids)
        num_accepted_str = (
            f"{num_accepted:,} design{'s' if num_accepted > 1 else ''} passing threshold ({fraction_accepted:.2%})"
            if num_accepted > 0
            else ":red[No designs passing threshold]"
        )
        threshold_str = threshold.format(descriptor.name)
        if threshold_str:
            threshold_labels.append(f"- **{threshold_str}**: {num_accepted_str}")
    st.markdown("\n".join(threshold_labels) or "No customizable thresholds available")



import copy
from pathlib import Path

import jsonschema


def validate_params(params: dict, schema: dict) -> None:
    """Validate input parameters against JSON schema.

    Args:
        params: Dictionary of input parameters.
        schema: JSON schema dictionary.

    Raises:
        jsonschema.ValidationError: If validation fails.
    """
    params = coerce_types(params, schema)
    jsonschema.validate(instance=params, schema=schema)


def coerce_types(params, schema):
    """Auto-convert parameter value types based on JSON schema."""

    # Remove None values, they are not passed to nextflow by scheduler
    params = {k: v for k, v in params.items() if v is not None}

    for key, value in params.items():
        if isinstance(value, Path):
            params[key] = str(value.resolve())
        if key in schema.get("properties", {}):
            prop = schema["properties"][key]
            if "type" in prop:
                target_type = prop["type"]
                if target_type == "integer":
                    try:
                        params[key] = int(value)
                    except (ValueError, TypeError):
                        pass
                elif target_type == "number":
                    try:
                        params[key] = float(value)
                    except (ValueError, TypeError):
                        pass
                elif target_type == "boolean":
                    if isinstance(value, str):
                        if value.lower() in ["true", "1", "yes"]:
                            params[key] = True
                        elif value.lower() in ["false", "0", "no"]:
                            params[key] = False
    return params


def flatten_schema(schema: dict) -> dict:
    """
    Flattens a JSON Schema that uses allOf + definitions into a single-level schema.
    """
    schema = copy.deepcopy(schema)  # don’t mutate original
    root = {
        "type": "object",
        "properties": {},
        "required": [],
        "additionalProperties": schema.get("additionalProperties", False),
    }

    for entry in schema.get("allOf", []):
        if "$ref" in entry:
            ref = entry["$ref"]
            # assumes refs are of the form "#/definitions/NAME"
            _, _, name = ref.partition("#/definitions/")
            definition = schema["definitions"][name]

            # merge properties
            props = definition.get("properties", {})
            root["properties"].update(props)

            # merge required keys
            root["required"].extend(definition.get("required", []))

    # deduplicate required
    root["required"] = list(dict.fromkeys(root["required"]))
    return root



import os
import re
from copy import deepcopy

import streamlit as st

from ovo.app.components.history_components import history_dropdown_component
from ovo.app.components.input_components import sequence_selection_fragment, initialize_workflow
from ovo.app.components.molstar_custom_component import molstar_custom_component, StructureVisualization
from ovo.app.components.navigation import show_prev_next_sections
from ovo.app.components.preview_components import visualize_rfdiffusion_preview
from ovo.app.components.scheduler_components import wait_with_statusbar
from ovo.app.components.submission_components import (
    pool_submission_inputs,
    review_workflow_submission,
    show_rfdiffusion_binder_seq_design_inputs,
)

from ovo.app.pages import jobs_page, designs_page, rf_binder_design_page
from ovo.app.utils.page_init import initialize_page
from ovo import (
    config,
    db,
    local_scheduler,
    storage,
    Design,
    get_username,
)
from ovo.core.database.models_rfdiffusion import (
    RFdiffusionBinderDesignWorkflow,
    RFdiffusionWorkflow,
    MODEL_WEIGHTS_BINDER,
)
from ovo.app.components.submission_components import show_rfdiffusion_advanced_settings
from ovo.core.database import descriptors_rfdiffusion, descriptors_refolding
from ovo.core.database.models import WorkflowTypes, Pool, DesignJob
from ovo.core.logic.design_logic_rfdiffusion import submit_rfdiffusion_preview
from ovo.core.utils.formatting import get_hashed_path_for_bytes, safe_filename
from ovo.core.utils.residue_selection import get_chains_and_contigs


@st.fragment
def intro_step():
    # Initialize the workflow object in session state
    initialize_workflow(
        page_key=__file__,
        workflow_name=RFdiffusionBinderDesignWorkflow.name,
        include_subclasses=True,
    )

    with st.container(width=850):
        st.markdown(
            f"""
This workflow enables starting from an existing binder-target complex, generating 
similar binder backbones and designing their sequences. It implements the end-to-end RFdiffusion partial diffusion
protocol,<sup>[2,3]</sup> where noise is added to the backbone of an existing binder chain.

Use cases:
- Generate binders with improved computational success rate, starting from a computationally successful design. <sup>[3]</sup>
- Generate binders with improved specificity and affinity for a target protein, using an existing binder as starting point. <sup>[2]</sup>
- Generate binders with altered specificity by starting from a binder–homolog complex (same family) with the target swapped for homolog. <sup>[2,3]</sup>
    <!--[2] also shows the success rate is better when altering the specificity than when starting de-novo. In [3] the starting design was not
                de novo from RFdiffusion but rather some modified structure using RF joint inpainting.  -->

This workflow noises and denoises the binder chain in the input structure for a specified number of timesteps (noising
strength) using RFdiffusion. This allows sampling backbones around the starting binder structure. The higher the number
of timesteps, the more the generated backbones will diverge from the original binder structure.

- **Input**
    - Protein structures — one or more PDB files with binder chain (A) and target chain (B), or OVO Design IDs
- **Main parameters**
    - Hotspot residues to select preferred binding site (optional, RFdiffusion may preserve the original binding site even without hotspots, but specifying hotspots explicitly enables computing hotspot-focused interface metrics) 
    - (Optional) Noising strength (num partial diffusion timesteps `partial_T`), the default is 20 out of 1-50
- **Output**
    - Designs — binder sequences and structures in complex with target protein
    - Design descriptors
        - Refolding metrics
        - Rosetta descriptors
        - Backbone metrics
        - Sequence composition

You can run this workflow as a next step after {rf_binder_design_page.title} — besides partial diffusion, the methods and computed 
metrics are the same.

### Features
                
View and filter your designs based on descriptors at {jobs_page.title}. All designs and descriptors are also available
for download. Most relevant descriptors are listed below:
   
#### Backbone metrics

- Computed for the RFdiffusion backbone design
- **Secondary structure composition**:
    - Percentages of helix, sheet, loop (the remaining residues); computed using PyDSSP
- Backbone-backbone contacts (Cα atoms within 8&nbsp;Å):
    - **{descriptors_rfdiffusion.N_CONTACTS_TO_INTERFACE.name}** — Number of pairwise contacts between all residues of binder and target
    - **{descriptors_rfdiffusion.N_CONTACTS_TO_HOTSPOTS.name}** — Number of pairwise contacts between the binder backbone and hotspot residues
    - **{descriptors_rfdiffusion.N_HOTSPOTS_ON_INTERFACE.name}** — Number of hotspot residues in contact with the binder backbone

#### Sequence composition

- Percentage of individual amino acids
- Charge at pH 7.4 and 5.5, isoelectric point
- Aromaticity
- And more; computed using BioPython ProteinAnalysis <sup>[5]</sup>

#### Rosetta descriptors <sup>[6]</sup>

- **{descriptors_rfdiffusion.PYROSETTA_DDG.name}** — Rosetta binding energy of the complex (in Rosetta energy units). Lower (negative) values indicate better binding affinity.
- **{descriptors_rfdiffusion.PYROSETTA_CMS.name}** — Rosetta molecular surface area (in square angstroms).
- **{descriptors_rfdiffusion.PYROSETTA_SAP_SCORE.name}** — Evaluates hydrophobicity of surface exposed regions. Positive values indicate hydrophobic molecules.

#### AlphaFold2 initial guess refolding metrics

- **{descriptors_refolding.AF2_PRIMARY_IPAE.name}** — Binder–target pose error (predicted) in angstroms. Predicted Aligned Error (residue–residue 
    matrix) averaged over binder→target residue values.
- **{descriptors_refolding.AF2_PRIMARY_TARGET_ALIGNED_BINDER_RMSD.name}** — Cα RMSD of the binder between RFdiffusion output and AF2 prediction (aligned on target)
- **{descriptors_refolding.AF2_PRIMARY_BINDER_PAE.name}** — Binder error (predicted) including the binder–target pose, in angstroms. Predicted Aligned Error 
    <sup>[4]</sup> (residue–residue matrix) averaged over binder→binder and binder→target residue values.

We also include AF2 binder pLDDT and pTM scores.
        
In RFdiffusion,<sup>[1]</sup> in silico success for binder design was defined as AF2&nbsp;Binder&nbsp;pLDDT >&nbsp;80,
AF2&nbsp;Interface&nbsp;PAE <&nbsp;10 and AF2&nbsp;monomer&nbsp;RMSD <&nbsp;1.0 (we do not include this metric with the
binder predicted in isolation, you could use AF2&nbsp;Binder&nbsp;RMSD which includes the pose RMSD in the complex).

These descriptors are computed for all designs. Additionally, you can run ProteinQC for the accepted designs at
{designs_page.title}, computing descriptors of predicted solubility, electrostatic and hydrophobic patch area, and more.

                
### Limitations
<!--Binder diversification is -->
- Only one target chain is supported. You may circumvent this by merging multiple chains into a single chain in your
input PDB file, adding residue index jumps (200) between the original chains, this influences both RFdiffusion and
AlphaFold2.
 <!---clipped to -32, 32 in alphafold2, colabdesign adds 50 between binder and target. How are inputs for proteinmpnn
 prepared? It could add 200 residues between the original chains and slow it down (a bit, is not quadratic..) --->

Request more features in our **TODO github issues link**!

### References
1. [RFdiffusion](https://www.nature.com/articles/s41586-023-06415-8) (and [Partial Diffusion GitHub guide](https://github.com/RosettaCommons/RFdiffusion?tab=readme-ov-file#partial-diffusion))
2. Partial diffusion, [Glögl et al. 2024](https://www.science.org/doi/full/10.1126/science.adp1779), Target-conditioned diffusion generates potent TNFR superfamily antagonists and agonists
3. Partial diffusion, [Vázquez Torres et al. 2023](https://www.nature.com/articles/s41586-023-06953-1), De novo design of high-affinity binders of bioactive helical peptides
4. [Predicted Aligned Error (PAE)](https://www.ebi.ac.uk/training/online/courses/alphafold/inputs-and-outputs/evaluating-alphafolds-predicted-structures-using-confidence-scores/pae-a-measure-of-global-confidence-in-alphafold-predictions/)
5. [BioPython ProteinAnalysis](https://biopython.org/docs/latest/api/Bio.SeqUtils.ProtParam.html#Bio.SeqUtils.ProtParam.ProteinAnalysis)
6. [PyRosetta descriptors](https://www.science.org/doi/10.1126/science.abn2100#supplementary-materials)
""",
            unsafe_allow_html=True,
        )


@st.fragment
def input_step():
    workflow: RFdiffusionBinderDesignWorkflow = st.session_state.workflows[__file__]

    left, _, right = st.columns([4, 1, 3])
    with left:
        st.subheader("Input structures")

        if not workflow.rfdiffusion_params.input_pdb_paths:
            st.markdown("##### Start from one or more previous designs")

            if initialize_workflow_input():
                # Input structure was just provided, generate contigs and rerun
                # workflow instance was updated in initialize_workflow(), read it again
                workflow = st.session_state.workflows[__file__]
                workflow.preview_job_id = None
                workflow.rfdiffusion_params.partial_diffusion = True
                workflow.rfdiffusion_params.timesteps = 20
                workflow.rfdiffusion_params.num_designs = 10
                # TODO use presets to initialize this instead
                workflow.refolding_params.primary_test = "af2_model_1_multimer_tt_3rec"

                # Generate a contig for each input file
                workflow.rfdiffusion_params.contigs = []
                for pdb_path in workflow.rfdiffusion_params.input_pdb_paths:
                    chain_contigs = get_chains_and_contigs(storage.read_file_str(pdb_path))
                    # read target length 100 from A1-100/0
                    chain_ids = sorted(chain_contigs.keys())
                    try:
                        assert chain_ids == ["A", "B"], (
                            f"This pipeline expects two chains - binder (A) and target (B), got: {chain_ids}"
                        )
                        assert chain_contigs["A"].removeprefix("A").startswith("1-"), (
                            f"Expected binder chain to start with residue 1, got: {chain_contigs['A']}"
                        )
                        binder_len = int(chain_contigs["A"].removeprefix("A").removeprefix("1-").removesuffix("/0"))
                        # 100-100/0 B20-130
                        workflow.rfdiffusion_params.contigs.append(f"{binder_len}-{binder_len}/0 {chain_contigs['B']}")
                    except Exception as e:
                        st.error(e)
                        workflow.rfdiffusion_params.input_name = None
                        workflow.rfdiffusion_params.input_pdb_paths = []
                        st.stop()
                        return

                st.rerun()

        elif st.button("Change input"):
            workflow.input_name = None
            workflow.rfdiffusion_params.input_pdb_paths = []
            st.rerun()

    with right:
        history_dropdown_component(
            page_key=__file__,
            workflow_name=RFdiffusionBinderDesignWorkflow.name,
            filter_func=lambda w: w.rfdiffusion_params.partial_diffusion,
        )

    if not workflow.rfdiffusion_params.input_pdb:
        # No input structure provided yet, stop here
        return

    st.subheader(workflow.input_name)

    if len(workflow.rfdiffusion_params.input_pdb_paths) > 1:
        indexes = {v: i for i, v in enumerate(workflow.rfdiffusion_params.input_pdb_paths, start=1)}
        input_pdb_path = st.selectbox(
            "Visualize input",
            options=workflow.rfdiffusion_params.input_pdb_paths,
            format_func=lambda x: f"{indexes[x]} | {x.split('/')[-1].removesuffix('.pdb')}",
        )
    else:
        input_pdb_path = workflow.rfdiffusion_params.input_pdb

    molstar_custom_component(
        structures=[StructureVisualization(pdb=storage.read_file_str(input_pdb_path))],
        height=500,
        key="diversification",
    )


def initialize_workflow_input() -> RFdiffusionBinderDesignWorkflow | None:
    workflow = st.session_state.workflows[__file__]

    design_ids_raw = st.text_area(
        "Design IDs to be diversified",
        placeholder="Enter design IDs separated by spaces, commas, or lines",
        value="\n".join(workflow.design_ids),
    )
    design_ids = [design_id.strip() for design_id in design_ids_raw.replace(",", " ").split()]

    if st.button("Confirm"):
        designs: list[Design] = db.select(Design, id__in=design_ids)
        if not designs:
            st.error("No designs found for the provided IDs.")
            return False
        if missing_ids := (set(design_ids).difference([d.id for d in designs])):
            st.error(f"Designs not found for IDs: {', '.join(missing_ids)}")
            return False
        pool_ids = sorted(set(design.pool_id for design in designs))

        # Get rfdiffusion params (use arbitrary pool in case of multiple pools)
        rfdiffusion_params = None
        source_workflow_names = set()
        for pool_id in pool_ids:
            pool = db.get(Pool, pool_id)
            if not pool.design_job_id:
                continue
            design_job = db.get(DesignJob, id=pool.design_job_id)
            if not design_job.workflow or not design_job.workflow.is_instance(RFdiffusionBinderDesignWorkflow):
                continue
            source_workflow_names.add(design_job.workflow.name)
            if rfdiffusion_params is None:
                rfdiffusion_params = deepcopy(design_job.workflow.rfdiffusion_params)

        assert source_workflow_names, "Designs do not come from RFdiffusion binder design workflow"
        assert len(source_workflow_names) == 1, (
            f"Expected all designs to come from the same workflow, got: {', '.join(source_workflow_names)}"
        )
        source_workflow_name = source_workflow_names.pop()
        if source_workflow_name != workflow.name:
            # Designs can come from a subclass of the current workflow, create appropriate workflow instance
            workflow = WorkflowTypes.get(source_workflow_name)()
            st.session_state.workflows[__file__] = workflow
        workflow.input_name = design_ids[0] if len(design_ids) == 1 else f"{len(design_ids)} designs"
        workflow.design_ids = design_ids
        workflow.rfdiffusion_params = rfdiffusion_params
        # Use ProteinMPNN path, the RFdiffusion PDB does not have any sidechain coords (even on the target chain)
        workflow.rfdiffusion_params.input_pdb_paths = [design.structure_path for design in designs]
        # Do not preserve the original hotspots - they are not correct anymore.
        # We could parse them from the standardized PDB but from which design? The first one?
        # We should just expose it to the user in the UI so they get visibility on it.
        workflow.rfdiffusion_params.hotspots = None

    st.markdown("##### Or upload a custom structure")

    uploaded_file = st.file_uploader(
        "Upload a PDB file with binder chain (A) and target chain (B)",
        type=["pdb", "pdb1"],
        key="input_pdb_file",
    )

    # Check if a file has just been uploaded
    if uploaded_file is not None:
        pdb_input_bytes = uploaded_file.getvalue()
        filename, _ = os.path.splitext(safe_filename(uploaded_file.name))
        workflow.input_name = filename
        workflow.rfdiffusion_params.input_pdb = storage.store_file_str(
            pdb_input_bytes.decode("utf-8"),
            f"project/{st.session_state.project.id}/inputs/{get_hashed_path_for_bytes(pdb_input_bytes)}/{workflow.input_name}.pdb",
            overwrite=False,
        )

    if workflow.rfdiffusion_params.input_pdb_paths:
        return workflow
    else:
        return None


@st.fragment()
def selection_step():
    workflow: RFdiffusionWorkflow = st.session_state.workflows[__file__]

    st.subheader("Select residues for partial diffusion")

    st.write(
        "Select binder residues that should be redesigned by applying RFdiffusion partial diffusion "
        "followed by ProteinMPNN sequence design. By default, we redesign the entire binder."
    )

    if not workflow.rfdiffusion_params.input_pdb:
        st.error("Please provide an input structure in the input structure step.")
        return

    if not workflow.rfdiffusion_params.contig:
        st.error("Please provide a contig in the previous step.")
        return

    if len(workflow.rfdiffusion_params.contigs) > 1:
        #
        # TODO support residue selection for multiple inputs
        #
        st.warning(
            "Selection of residues is not yet supported when multiple input structures are provided. "
            "Whole binder will be redesigned."
        )
        return

    sequence_selection_fragment(
        __file__, workflow.input_name, write_segments=False, contig_index=0, partial_diffusion=True
    )


@st.fragment()
def preview_step():
    workflow: RFdiffusionWorkflow = st.session_state.workflows[__file__]
    st.subheader("Preview design")

    if not workflow.rfdiffusion_params.input_pdb:
        st.error("Please provide an input structure in the input structure step.")
        return

    st.write(f"Using contig: {workflow.rfdiffusion_params.contig}")

    # Generate preview
    st.write("#### Generate preview")
    num_timesteps = 10
    with st.columns([2, 1])[0]:
        st.write(f"""
        Generate a quick RFdiffusion preview of the design with {num_timesteps} partial diffusion timesteps to verify your inputs.

        This should take 2-10 minutes depending on the length of the target and binder.
        """)

    if st.button(":material/wand_stars: Generate preview"):
        with st.spinner("Submitting RFdiffusion job..."):
            workflow.preview_job_id = submit_rfdiffusion_preview(
                workflow, partial_diffusion=True, timesteps=num_timesteps
            )

    # Check if needed parameters are set
    if not workflow.preview_job_id:
        return

    # Wait until task is done
    result = wait_with_statusbar(local_scheduler, workflow.preview_job_id, label="Running RFdiffusion...")

    # Exit if job failed
    if not result:
        return

    visualize_rfdiffusion_preview(workflow, output_dir=local_scheduler.get_output_dir(workflow.preview_job_id))


@st.fragment()
def settings_step():
    st.subheader("Settings")

    workflow: RFdiffusionWorkflow = st.session_state.workflows[__file__]

    if not workflow.rfdiffusion_params.input_pdb:
        st.error("Please provide an input structure in the input structure step.")
        return

    pool_submission_inputs(__file__)

    with st.columns([1, 2])[0]:
        is_admin = get_username() in config.auth.admin_users
        workflow.rfdiffusion_params.num_designs = st.number_input(
            "Number of structure designs (RFdiffusion backbones) per each input structure",
            min_value=1,
            max_value=config.props.rfdiffusion_backbones_limit_admin
            if is_admin
            else config.props.rfdiffusion_backbones_limit,
            value=workflow.rfdiffusion_params.num_designs,
            key="num_designs",
        )

    show_rfdiffusion_binder_seq_design_inputs(workflow)

    with st.columns([1, 2])[0]:
        workflow.rfdiffusion_params.model_weights = st.selectbox(
            "Model weights",
            help="Use 'active site' model weights to hold better selected residues specified in the contig.",
            index=MODEL_WEIGHTS_BINDER.index(workflow.rfdiffusion_params.model_weights)
            if workflow.rfdiffusion_params.model_weights
            else 0,
            key="active_site",
            options=MODEL_WEIGHTS_BINDER,
        )

    with st.columns([1, 2])[0]:
        workflow.rfdiffusion_params.hotspots = st.text_input(
            "Hotspot residues (optional)",
            placeholder="For example B123,B124,B131",
            value=workflow.rfdiffusion_params.hotspots,
            key="hotspots",
        )
        if workflow.rfdiffusion_params.hotspots:
            if not all(
                re.fullmatch("[A-Z][0-9]+", hotspot) for hotspot in workflow.rfdiffusion_params.hotspots.split(",")
            ):
                st.error("Invalid hotspots format, expected 'B123,B124,B131'")

    # TODO carry over hotspots from original design if applicable
    if not workflow.rfdiffusion_params.hotspots:
        st.write("""
        :material/info: Note that any hotspots used in the original design were not carried over.
        Hotspots should be specified in the context of the designed structure (binder in chain A, target in chain B).
        """)

    show_rfdiffusion_advanced_settings(workflow)


@st.fragment
def review_step():
    st.subheader("Review settings")

    review_workflow_submission(__file__)


if __name__ == "__main__":
    initialize_page(page_title="Binder diversification")

    show_prev_next_sections(
        key=__file__,
        title="♻️ Binder diversification",
        sections={
            "intro": intro_step,
            "input": input_step,
            "selection": selection_step,
            "preview": preview_step,
            "settings": settings_step,
            "review": review_step,
        },
    )



import random


class ColorPicker:
    def __init__(self, seed: int = 27):
        self.rng = random.Random(seed)
        random.seed(seed)
        self.assignedColorsCount = 0
        # modified tab10 palette
        self.prefferedColors = [
            "#1f77b4",
            "#ff7f0e",
            "#2ca02c",
            "#d62728",
            "#9467bd",
            "#9e4634",
            "#e377c2",
            "#bcbd22",
            "#17becf",
            "#6f6f6f",
        ]
        self.assignedValues = []
        self.assignedColorsList = []

    def __call__(self, value):
        if value in self.assignedValues:
            idx = self.assignedValues.index(value)
            return self.assignedColorsList[idx]

        if self.assignedColorsCount < len(self.prefferedColors):
            color = self.prefferedColors[self.assignedColorsCount]
        else:
            r = lambda: self.rng.randint(0, 255)
            color = "#%02X%02X%02X" % (r(), r(), r())

        self.assignedColorsCount += 1
        self.assignedColorsList.append(color)
        self.assignedValues.append(value)
        return color


def mix_colors(first_hex, second_hex):
    return "#" + "".join(
        f"{(int(first_hex[i : i + 2], 16) + int(second_hex[i : i + 2], 16)) // 2:02x}" for i in (1, 3, 5)
    )



from collections import Counter
from ovo.core.database.models import StructureFileDescriptor
from ovo.core.plugins import load_variable, plugin_descriptors

# Add descriptors and presets from plugins
# key -> { label: str, x: Descriptor, y: Descriptor }
ALL_DESCRIPTORS = []
ALL_DESCRIPTORS_BY_KEY = {}
PRESETS = {}
for descriptor_path in [
    "ovo.core.database.descriptors_refolding",
    "ovo.core.database.descriptors_rfdiffusion",
    "ovo.core.database.descriptors_bindcraft",
    "ovo.core.database.descriptors_proteinqc",
] + plugin_descriptors:
    module_name = descriptor_path.split(".")[0]
    # Load DESCRIPTORS from plugin and extend ALL_DESCRIPTORS
    _descriptors = load_variable(f"{descriptor_path}:DESCRIPTORS")
    assert isinstance(_descriptors, list), "Expected DESCRIPTORS to be a list, got {} in {}".format(
        type(_descriptors), descriptor_path
    )
    for descriptor in _descriptors:
        if descriptor.key in ALL_DESCRIPTORS:
            raise ValueError(f"Duplicate descriptor key '{descriptor.key}' found in {descriptor_path}")
        ALL_DESCRIPTORS.append(descriptor)
        ALL_DESCRIPTORS_BY_KEY[descriptor.key] = descriptor
    try:
        # Load PRESETS from plugin and update PRESETS
        _presets = load_variable(f"{descriptor_path}:PRESETS")
        assert isinstance(_presets, list), "Expected PRESETS to be list, got {} in {}".format(
            type(_presets), descriptor_path
        )
        for preset in _presets:
            assert "label" in preset and "x" in preset and "y" in preset, (
                "Expected preset to have 'label', 'x', and 'y' fields, got {}".format(preset)
            )
            label = preset["label"]
            PRESETS[f"{descriptor_path} {label}"] = preset
    except AttributeError:
        # No presets defined in the plugin
        pass

ALL_DESCRIPTOR_KEYS = [descriptor.key for descriptor in ALL_DESCRIPTORS]
ALL_DESCRIPTOR_KEYS_SET = frozenset(ALL_DESCRIPTOR_KEYS)
# NumericGlobalDescriptor keys should be unique
if len(ALL_DESCRIPTOR_KEYS) != len(ALL_DESCRIPTOR_KEYS_SET):
    counts = Counter(ALL_DESCRIPTOR_KEYS)
    duplicates = {key: value for key, value in counts.items() if value > 1}
    raise ValueError(f"Duplicate descriptor keys found: {duplicates}")
# Descriptor keys should contain two pipes ("|")
for key in ALL_DESCRIPTOR_KEYS:
    if not key:
        raise ValueError(f"Some descriptors are missing a key: {[d for d in ALL_DESCRIPTORS if not d.key]}")
    if key.count("|") != 2:
        raise ValueError(f"Descriptor key should contain two | separators (pipeline|tool|descriptor), found: {key}")

STRUCTURE_PATH_DESCRIPTORS = [d for d in ALL_DESCRIPTORS if isinstance(d, StructureFileDescriptor)]
SEQUENCE_DESIGN_PATH_DESCRIPTORS = [d for d in STRUCTURE_PATH_DESCRIPTORS if d.structure_type == "sequence_design"]



from dataclasses import dataclass
from functools import reduce

import pandas as pd
import plotly.graph_objects as go
import streamlit as st
from plotly import express as px
from ovo.app.utils.cached_db import (
    get_cached_available_descriptors,
    get_cached_descriptor_values,
    get_cached_designs_accept_field,
)
from ovo.core.database import Threshold, NumericGlobalDescriptor, Descriptor
from ovo.core.database.descriptors import PRESETS


@dataclass
class PlotSettings:
    x: Descriptor | None = None
    y: Descriptor | None = None
    color: Descriptor | None = None
    color_type: str = "range"

    def get_x_key(self):
        return self.x.key if self.x else None

    def get_y_key(self):
        return self.y.key if self.y else None

    @classmethod
    def from_query_params(cls, descriptors_by_key: dict[str, Descriptor]):
        for field in ["x", "y", "color"]:
            # Handle case when descriptor is not available anymore - such as when switching between workflows
            if field in st.query_params and st.query_params[field] not in descriptors_by_key:
                st.warning(f"Descriptor not available: {st.query_params[field]}")
                st.query_params.pop(field)
        return cls(
            x=descriptors_by_key[st.query_params["x"]] if "x" in st.query_params else None,
            y=descriptors_by_key[st.query_params["y"]] if "y" in st.query_params else None,
            color=descriptors_by_key[st.query_params["color"]] if "color" in st.query_params else None,
            color_type=st.query_params["color_type"] if "color_type" in st.query_params else "range",
        )

    def update_query_params(self):
        st.query_params["color_type"] = self.color_type
        for field in ["x", "y", "color"]:
            if getattr(self, field):
                st.query_params[field] = getattr(self, field).key
            else:
                st.query_params.pop(field, None)


def descriptor_scatterplot_input_component(design_ids: list[str]) -> PlotSettings | None:
    descriptors_by_key = get_cached_available_descriptors(design_ids)

    # select only global numeric descriptors
    descriptors_by_key = {k: d for k, d in descriptors_by_key.items() if isinstance(d, NumericGlobalDescriptor)}

    descriptor_options = list(descriptors_by_key.keys())

    settings = PlotSettings.from_query_params(descriptors_by_key)

    available_presets = {
        key: value
        for key, value in PRESETS.items()
        if value["x"].key in descriptors_by_key and value["y"].key in descriptors_by_key
    }

    if st.session_state.get("scatterplot_update"):
        # update triggered from another component
        update, value = st.session_state["scatterplot_update"]
        if update == "x":
            st.session_state["scatterplot_x"] = value
        elif update == "y":
            st.session_state["scatterplot_y"] = value
        else:
            print("Ignoring unknown update", update, value)

    # Update settings if user just selected something in one of the dropdowns
    if st.session_state.get("scatterplot_preset") or (not settings.x and not settings.y and available_presets):
        if st.session_state.get("scatterplot_preset"):
            # user has just selected a preset
            new_preset = st.session_state.get("scatterplot_preset")
            st.session_state["flash_preset"] = new_preset
        else:
            # No x or y axis selected, choose first preset
            new_preset = list(available_presets.keys())[0]
        settings.x = available_presets[new_preset]["x"]
        settings.y = available_presets[new_preset]["y"]
        st.session_state["scatterplot_x"] = settings.x.key
        st.session_state["scatterplot_y"] = settings.y.key
        st.session_state["scatterplot_preset"] = None  # clear preset dropdown
        settings.update_query_params()
        st.rerun()
    elif st.session_state.get("scatterplot_x") and settings.get_x_key() != st.session_state.get("scatterplot_x"):
        # user has just changed the X axis
        # note that the session state is already updated here if user has just changed the value of the dropdown below
        settings.x = descriptors_by_key[st.session_state.get("scatterplot_x")]
        settings.update_query_params()
    elif st.session_state.get("scatterplot_y") and settings.get_y_key() != st.session_state.get("scatterplot_y"):
        # user has just changed the Y axis
        # note that the session state is already updated here if user has just changed the value of the dropdown below
        settings.y = descriptors_by_key[st.session_state.get("scatterplot_y")]
        settings.update_query_params()

    left, right = st.columns([1.1, 2], gap="medium", vertical_alignment="bottom")
    with left:
        st.selectbox(
            "Preset",
            placeholder="One preset available"
            if len(available_presets) == 1
            else f"{len(available_presets)} presets available",
            format_func=lambda key: PRESETS[key]["label"],
            options=list(available_presets.keys()),
            key="scatterplot_preset",
            index=None,
        )
    with right:
        if st.session_state.get("flash_preset"):
            st.success(f"Preset selected: {st.session_state['flash_preset']}")
            del st.session_state["flash_preset"]

    x_col, y_col, _ = st.columns([1, 1, 0.5], gap="medium")
    with x_col:
        st.selectbox(
            "X axis",
            format_func=lambda descriptor_key: format_descriptor_name(descriptors_by_key[descriptor_key]),
            options=descriptor_options,
            key="scatterplot_x",
            index=None,
        )
        if settings.x and settings.x.description:
            st.caption(settings.x.description)

    with y_col:
        st.selectbox(
            "Y axis",
            format_func=lambda descriptor_key: format_descriptor_name(descriptors_by_key[descriptor_key]),
            options=descriptor_options,
            key="scatterplot_y",
            index=None,
        )
        if settings.y and settings.y.description:
            st.caption(settings.y.description)

    # with color_col:
    #     key = 'scatterplot_color'
    #     color_descriptor = get_descriptor_by_key(descriptors, key) or Descriptor(name='run_id', key='run_id')
    #     color = st.selectbox(
    #         "Color",
    #         descriptors,
    #         format_func=lambda descriptor: (
    #             f"{descriptor.name} ({descriptor.tool})" if descriptor.tool else descriptor.name
    #         ),
    #         index=get_descriptor_index(descriptors, color_descriptor),
    #     )
    #     st.query_params[key] = color.key
    #
    # with color_type_col:
    #     color_type = st.selectbox(
    #         "Color type", ["range", "PDB density"], index=0, disabled=True
    #     )

    return settings


def descriptor_scatterplot_design_explorer_component(settings: PlotSettings, design_ids: list[str]) -> list[str]:
    if not settings.x or not settings.y:
        st.caption("Please select the X and Y axes to see the scatterplot.")
        return design_ids

    x_values = get_cached_descriptor_values(settings.x.key, design_ids=design_ids)
    y_values = get_cached_descriptor_values(settings.y.key, design_ids=design_ids)
    # FIXME
    color_values = pd.Series([design_id.split("_")[1] for design_id in design_ids], index=design_ids)

    values_by_name = {
        settings.x.name: x_values,
        settings.y.name: y_values,
    }
    print_missing(values_by_name)

    st.markdown(
        """
        <style>
        /* center the plot */
        .js-plotly-plot .plotly .svg-container { margin: 0 auto; }
        </style>
        """,
        unsafe_allow_html=True,
    )
    scatterplot_df = pd.DataFrame({"x": x_values, "y": y_values, "color": color_values, "design_id": design_ids})
    fig = px.scatter(
        scatterplot_df,
        x="x",
        y="y",
        labels={"x": settings.x.name, "y": settings.y.name},
        range_x=settings.x.get_plot_range(x_values),
        range_y=settings.y.get_plot_range(y_values),
        width=800,
        height=600,
        color="color",
        custom_data=["design_id"],
        marginal_x="histogram",
        marginal_y="histogram",
    )

    fig.update_layout(dragmode="select")
    event = st.plotly_chart(fig, on_select="rerun", width="content", key="scatterplot")

    # descriptor to create a subpool from lasso/box selected designs
    if not event["selection"]["box"] and not event["selection"]["lasso"]:
        # No selection yet
        st.info(
            "Click and drag to select a region in the plot.",
            icon=":material/info:",
        )
        return design_ids
    else:
        selected_design_ids = [point["customdata"][0] for point in event["selection"]["points"]]
        if box := event["selection"]["box"]:
            x_min, x_max = get_trimmed_min_max(box[0]["x"], settings.x)
            y_min, y_max = get_trimmed_min_max(box[0]["y"], settings.y)
            selection_label = (
                f"with **{format_range(x_min, x_max, settings.x)}** and **{format_range(y_min, y_max, settings.y)}**"
            )
        else:
            selection_label = "using lasso selection"
        st.info(
            f"Selected {len(selected_design_ids):,} designs {selection_label}. "
            f"Double-click on the plot to clear the selection.",
            icon=":material/info:",
        )
        return selected_design_ids


def descriptor_scatterplot_pool_details_component(
    settings: PlotSettings,
    design_ids: list[str],
    selected_thresholds: dict[str, Threshold] = None,
    highlight_accepted: bool = False,
) -> tuple[list[str], str | None]:
    if not settings.x or not settings.y:
        return design_ids, None

    x_values = get_cached_descriptor_values(settings.x.key, design_ids=design_ids)
    y_values = get_cached_descriptor_values(settings.y.key, design_ids=design_ids)
    # FIXME
    color_values = pd.Series([design_id.split("_")[1] for design_id in design_ids], index=design_ids)

    values_by_name = {
        settings.x.name: x_values,
        settings.y.name: y_values,
    }
    print_missing(values_by_name)

    st.markdown(
        """
        <style>
        /* center the plot */
        .js-plotly-plot .plotly .svg-container { margin: 0 auto; }
        </style>
        """,
        unsafe_allow_html=True,
    )
    scatterplot_df = pd.DataFrame({"x": x_values, "y": y_values, "color": color_values, "design_id": design_ids})
    if highlight_accepted:
        accepted_values = get_cached_designs_accept_field(design_ids)
        scatterplot_df["accepted"] = accepted_values

        # FIXME: When we use continuous column this will need to be changed
        color_scale = px.colors.qualitative.Plotly
        unique_colors = scatterplot_df["color"].unique()
        color_map = {value: color_scale[i % len(color_scale)] for i, value in enumerate(unique_colors)}
        scatterplot_df["color_mapped"] = scatterplot_df["color"].map(color_map)

        # Hover template
        # TODO: Use X and Y names in the hover
        hovertemplate = (
            "<b>%{customdata[0]}</b><br>"
            "pool=%{customdata[1]}<br>"
            f"{settings.x.name}=%{{x}}<br>"
            f"{settings.y.name}=%{{y}}<br>"
            "<extra></extra>"  # Remove trace name from hover
        )

        # Plot seperately accepted and rejected points
        accepted_df = scatterplot_df[scatterplot_df["accepted"] == True]
        rejected_df = scatterplot_df[scatterplot_df["accepted"] == False]

        rejected_trace = go.Scatter(
            x=rejected_df["x"],
            y=rejected_df["y"],
            mode="markers",
            marker=dict(
                color=rejected_df["color_mapped"],
            ),
            customdata=rejected_df[["design_id", "color"]],
            legendgroup="rejected",
            showlegend=False,
            hovertemplate=hovertemplate,
        )
        accepted_trace = go.Scatter(
            x=accepted_df["x"],
            y=accepted_df["y"],
            mode="markers",
            marker=dict(
                color=accepted_df["color_mapped"],
                line=dict(color="lime", width=2),
            ),
            customdata=accepted_df[["design_id", "color"]],
            legendgroup="accepted",
            showlegend=False,
            hovertemplate=hovertemplate,
        )
        layout = go.Layout(
            xaxis=dict(title=settings.x.name, range=settings.x.get_plot_range(scatterplot_df["x"])),
            yaxis=dict(title=settings.y.name, range=settings.y.get_plot_range(scatterplot_df["y"])),
            width=800,
            height=600,
        )

        # FIXME: When we use continuous column this will need to be changed
        # Create individual legend traces for each color (category)
        legend_traces = []
        for category, color in color_map.items():
            legend_trace = go.Scatter(
                x=[None],
                y=[None],  # Invisible data points
                mode="markers",
                name=category,
                marker=dict(
                    color=color,
                    size=10,  # Marker size in the legend
                ),
                legendgroup=category,  # Grouping by category to ensure color consistency in the legend
                showlegend=True,  # This ensures the legend is shown
            )
            legend_traces.append(legend_trace)

        fig = go.Figure(
            data=[rejected_trace, accepted_trace] + legend_traces,
            layout=layout,
        )

    else:
        fig = px.scatter(
            scatterplot_df,
            x="x",
            y="y",
            labels={"x": settings.x.name, "y": settings.y.name},
            range_x=settings.x.get_plot_range(x_values),
            range_y=settings.y.get_plot_range(y_values),
            width=800,
            height=600,
            color="color",
            custom_data=["design_id"],
        )

    # Add a rectangle for reference based on the selected thresholds
    if selected_thresholds and (settings.x.key in selected_thresholds or settings.y.key in selected_thresholds):
        x0, x1, xref = -0.04, -0.005, "paper"  # mark a small rectangle on the left
        y0, y1, yref = -0.07, -0.01, "paper"  # mark a small rectangle on the bottom
        if settings.x.key in selected_thresholds and selected_thresholds[settings.x.key].enabled:
            t = selected_thresholds[settings.x.key]
            x0, x1 = t.min_value, t.max_value
            xref = None
            # offset range for integer values to make it clear that the range is inclusive
            is_int = all(int(v) == v for v in x_values if not pd.isna(v))
            if is_int and x0 is not None:
                x0 -= 0.5
            if is_int and x1 is not None:
                x1 += 0.5
        if settings.y.key in selected_thresholds and selected_thresholds[settings.y.key].enabled:
            t = selected_thresholds[settings.y.key]
            y0, y1 = t.min_value, t.max_value
            yref = None
            # offset range for integer values to make it clear that the range is inclusive
            is_int = all(int(v) == v for v in y_values if not pd.isna(v))
            if is_int and y0 is not None:
                y0 -= 0.5
            if is_int and y1 is not None:
                y1 += 0.5
        if xref is None or yref is None:
            # use axis limits if specified, otherwise use actual min/max values
            (xlim0, xlim1) = fig.layout.xaxis.range or (x_values.min(), x_values.max())
            (ylim0, ylim1) = fig.layout.yaxis.range or (y_values.min(), y_values.max())
            fig.add_shape(
                type="rect",
                x0=x0 if x0 is not None else xlim0,
                x1=x1 if x1 is not None else xlim1,
                xref=xref,
                y0=y0 if y0 is not None else ylim0,
                y1=y1 if y1 is not None else ylim1,
                yref=yref,
                fillcolor="#00ff00",
                opacity=0.15,
                line_width=0,
            )

    fig.update_layout(dragmode="select")
    event = st.plotly_chart(fig, on_select="rerun", width="content", key="scatterplot")

    if not event["selection"]["box"] and not event["selection"]["lasso"]:
        return design_ids, None
    else:
        selected_design_ids = [point["customdata"][0] for point in event["selection"]["points"]]
        if box := event["selection"]["box"]:
            x_min, x_max = get_trimmed_min_max(box[0]["x"], settings.x)
            y_min, y_max = get_trimmed_min_max(box[0]["y"], settings.y)
            selection_label = (
                f"**{format_range(x_min, x_max, settings.x)}** and **{format_range(y_min, y_max, settings.y)}**"
            )
        else:
            selection_label = "Using lasso selection"
        return selected_design_ids, selection_label


def format_descriptor_name(descriptor: Descriptor) -> str:
    return f"{descriptor.name} ({descriptor.tool})" if descriptor.tool else descriptor.name


def get_trimmed_min_max(box_selection, descriptor):
    """Crop selected range to actually possible descriptor range - return None if value is outside of range"""
    min_value, max_value = min(box_selection), max(box_selection)
    if descriptor.min_value is not None and min_value < descriptor.min_value:
        min_value = None
    if descriptor.max_value is not None and max_value > descriptor.max_value:
        max_value = None
    return min_value, max_value


def format_range(min_value, max_value, descriptor):
    if min_value is not None and max_value is not None:
        return f"{descriptor.name} within {min_value:.2f} — {max_value:.2f}"
    elif min_value is not None:
        return f"{descriptor.name} ≥ {min_value:.2f}"
    elif max_value is not None:
        return f"{descriptor.name} ≤ {max_value:.2f}"


def print_missing(values_by_name: dict[str, pd.Series]):
    num_missing_by_name = {}
    for name, values in values_by_name.items():
        if num_missing := values.isna().sum():
            num_missing_by_name[name] = num_missing

    if not num_missing_by_name:
        return
    num_total = len(next(iter(values_by_name.values())))
    num_missing = reduce(lambda x, y: x | y, [values.isna() for values in values_by_name.values()]).sum()
    if num_missing == num_total:
        st.warning("No designs available for selected combination of descriptors")
    elif len(set(num_missing_by_name.values())) == 1:
        # all missing counts are the same
        st.warning(
            f"Not showing {num_missing:,}/{num_total:,} designs with missing "
            + " & ".join(f"**{name}**" for name in num_missing_by_name.keys())
        )
    else:
        st.warning(
            f"Not showing {num_missing:,}/{num_total:,} designs with missing "
            + " or ".join(f"**{name}** ({count:,} missing)" for name, count in num_missing_by_name.items())
        )



import typer
from rich.panel import Panel
from ovo import console, __version__
from ovo.cli import init
from ovo.cli import module
from ovo.cli import app as app_cli
from ovo.cli import scheduler_cli
from ovo.cli.common import OVOCliError


app = typer.Typer(
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)

app.add_typer(init.app, name="init")
app.add_typer(module.app, name="")
app.add_typer(app_cli.app, name="")
app.add_typer(scheduler_cli.app, name="scheduler")


def main():
    version = f"Version [green]{__version__}[/green]"
    console.print(f"""
   ▄▀▀█▄  ▄▖   ▄▄  ▄▀▀█▄   
  █    █▄ █▌   ██ █▄▀▄▀█▄  
 █     ██  █▌ ██ █     ██  
  ▀▄▄▄█▀    ███   ▀▄▄▄█▀   
 {version.rjust(39)}
    """)

    try:
        app()
    except OVOCliError as e:
        console.print(Panel.fit(str(e), title="CLI Error", border_style="red"))


if __name__ == "__main__":
    main()



from cmap import Colormap
from ovo.core.utils.formatting import get_hash_of_bytes
import matplotlib.colors as mc
import colorsys


def get_color_from_str(value: str, colormap_name: str) -> str:
    # Colormaps: https://cmap-docs.readthedocs.io/en/latest/catalog/
    cmap = Colormap(colormap_name)

    hash_str = get_hash_of_bytes(value.encode())

    # get number between 0.0-1.0
    hash_number = (int(hash_str[:5], 16) % 1000) / 1000

    return cmap(hash_number).hex.replace("#", "0x").lower()


def darken_lighten(hex_color, factor: float):
    """
    Darken or lighten a hex color by multiplying its lightness component in HLS space.
    factor < 1 → darker, factor > 1 → lighter
    """
    rgb = mc.to_rgb(hex_color)
    h, l, s = colorsys.rgb_to_hls(*rgb)
    l = max(0, min(1, l * factor))
    r, g, b = colorsys.hls_to_rgb(h, l, s)
    return mc.to_hex((r, g, b))


def lighten(hex_color, factor=1.3):
    return darken_lighten(hex_color, factor=factor)


def darken(hex_color, factor=1.3):
    return darken_lighten(hex_color, factor=1 / factor)



import streamlit as st

from ovo.app.utils.cached_db import get_cached_project_ids_and_names
from ovo.core.logic.project_logic import get_or_create_personal_project
from ovo.core.logic.user_settings_logic import get_or_create_user_settings
from ovo import db, get_username, config
from ovo.core.database.models import Project


@st.fragment()
@st.dialog("Input new project name")
def create_project_dialog():
    author = get_username()
    project_name = st.text_input("Project name", placeholder="My Project")
    public = st.toggle(label="Public", value=True)

    st.caption(
        "Public projects are visible to all users with access to this app. "
        "Private projects are only visible to you"
        + (
            ", but are accessible by other users when shared via a link."
            if config.auth.allow_private_project_link_access
            else "."
        )
    )

    if st.button("Create project"):
        if project_name == "":
            st.error("Please provide a project name.")
            return

        if db.count(Project, name=project_name, public=True):
            st.error(f'Public project "{project_name}" already exists. Please choose another name.')
            return
        elif db.count(Project, name=project_name, author=author):
            st.error(f'You already have a private project named "{project_name}". Please choose another name.')
            return

        new_project = Project(name=project_name, author=author, public=public)
        db.save(new_project)

        st.session_state.new_project_name = project_name
        st.session_state.project = new_project

        st.rerun()
        return


def get_query_arg_project() -> Project | None:
    if project_id := st.query_params.get("project_id"):
        if not config.auth.allow_private_project_link_access and project_id not in get_cached_project_ids_and_names(
            username=get_username()
        ):
            st.error(f"Project URL **{project_id}** not accessible, redirecting to last project.")
            del st.query_params["project_id"]
            return None
        else:
            return db.get(Project, id=project_id)
    return None


def project_sidebar_component():
    dropdown_key = "project_dropdown"
    success_message = None

    if st.session_state.project is None:
        if project := get_query_arg_project():
            # Load project from URL query parameter
            success_message = f"Opened project from URL"
            st.session_state.project = project
        elif config.props.read_only:
            # In read-only mode, just select the first available project
            project_ids_and_names = get_cached_project_ids_and_names(username=get_username())
            if not project_ids_and_names:
                st.error(f"No projects available to user {get_username()} in read-only mode.")
                st.stop()
            st.session_state.project = db.get(Project, id=list(project_ids_and_names.keys())[0])
        else:
            # Restore last selected project or select personal project
            user_settings = get_or_create_user_settings()
            if user_settings.last_project_id and (project := db.get(Project, user_settings.last_project_id)):
                success_message = f"Resuming in project **{project.name}**"
                st.session_state.project = project
            else:
                st.session_state.project = get_or_create_personal_project()

    project_ids_and_names = get_cached_project_ids_and_names(
        username=get_username(), extra_project_ids=[st.session_state.project.id]
    )
    project_ids = sorted(project_ids_and_names.keys(), key=lambda x: project_ids_and_names[x].lower())

    # Get dropdown value from session ID directly (or from session_state.project when dropdown is not yet rendered)
    selected_project_id = st.session_state.get(dropdown_key, st.session_state.project.id)

    if st.session_state.new_project_name:
        # Select project if it was just created
        success_message = f'Project "**{st.session_state.new_project_name.strip()}**" created'
        st.session_state.new_project_name = None
        # force selecting the new project in the dropdown
        selected_project_id = st.session_state.project.id

    st.sidebar.selectbox(
        "**Project**",
        format_func=project_ids_and_names.get,
        key=dropdown_key,
        options=project_ids,
        index=project_ids.index(selected_project_id),
    )

    # Hack needed to hide tooltips because they disrupt clicking on the dropdown item (as of Dec 2024)
    dropdown_style = """
    <style>
    .stTooltipContent {
    pointer-events: none;
    }
    </style>
    """
    st.markdown(dropdown_style, unsafe_allow_html=True)

    if selected_project_id != st.session_state.project.id:
        # Update selected project when changed
        selected_project = db.get(Project, id=selected_project_id)
        success_message = f"Selected project **{selected_project.name}**"
        st.session_state.project = selected_project
        if not config.props.read_only:
            user_settings = get_or_create_user_settings()
            user_settings.last_project_id = selected_project.id
            db.save(user_settings)

    if selected_project_id != st.query_params.get("project_id"):
        st.query_params["project_id"] = selected_project_id

    if config.props.read_only:
        st.sidebar.info("Read-only mode")
    elif st.sidebar.button(":material/add: Create new project"):
        create_project_dialog()

    if success_message:
        st.sidebar.success(success_message)



from abc import ABC
from datetime import datetime
import time


class SchedulerTypes:
    REGISTERED_CLASSES = {}

    @classmethod
    def register(cls, name=None):
        assert name is None or isinstance(name, str), (
            "Usage: SchedulerTypes.register() or SchedulerTypes.register(name='CustomName')"
        )

        def decorator(registered_class):
            cls.REGISTERED_CLASSES[name or registered_class.__name__] = registered_class
            return registered_class

        return decorator


class JobNotFound(Exception):
    pass


class Scheduler(ABC):
    # special case values for imported jobs
    IMPORTED_SCHEDULER_KEY: str = "imported"
    IMPORTED_JOB_ID: str = "imported"

    def __init__(
        self, name: str, workdir: str, reference_files_dir: str, allow_submit: bool = True, submission_args: dict = None
    ):
        """
        Args:
            name: Human-readable label for this scheduler
            workdir: Absolute path to working directory for nextflow workflow execution (local filesystem path or S3 URI)
            reference_files_dir: Absolute path to directory with model weights and other reference files
            allow_submit: Whether to allow job submission
            submission_args: Default submission arguments, can be overriden in submit method
        """
        self.name = name
        self.workdir = workdir
        self.reference_files_dir = reference_files_dir
        self.allow_submit = allow_submit
        self.submission_args = submission_args or {}

    def submit(self, pipeline_name: str, params: dict = None, submission_args: dict = None) -> str:
        """
        Submits a workflow asynchronously and returns the PID.

        :param pipeline_name: Workflow name and revision (ovo.rfdiffusion-end-to-end or github url with @version)
        :param params: Dictionary of parameters to pass to the workflow.
        :param submission_args: Submission arguments for the scheduler, overrides values in self.submission_args
        :return: Scheduler job ID
        """
        raise NotImplementedError()

    def get_status_label(self, job_id: str) -> str:
        """Get human-readable job status label. Should NOT be used to determine job status."""
        raise NotImplementedError()

    def get_result(self, job_id: str) -> bool | None:
        """Get job result: True if successful, False if failed, None if still running."""
        raise NotImplementedError()

    def get_log(self, job_id: str) -> str | None:
        """Get job execution log"""
        raise NotImplementedError()

    def cancel(self, job_id):
        """Cancel job execution"""
        raise NotImplementedError()

    def get_output_dir(self, job_id: str):
        """Get job output path"""
        raise NotImplementedError()

    def get_job_start_time(self, job_id: str) -> datetime | None:
        """Get job start time"""
        raise NotImplementedError()

    def get_job_stop_time(self, job_id: str) -> datetime | None:
        """Get job end time"""
        raise NotImplementedError()

    def get_startup_time_minutes(self):
        """Get startup time of a task (in minutes)"""
        return 0

    def wait(self, job_id: str, timeout: int = None, interval: int | float = 10) -> bool:
        """
        Wait synchronously using sleep loop for job completion, with optional timeout.

        :param job_id: Scheduler job ID
        :param timeout: Maximum time to wait in seconds, None for no timeout
        :param interval: Interval between status checks in seconds
        :return: True if successful, False if failed, raises TimeoutError if timed out
        """
        start_time = time.time()
        while self.get_result(job_id) is None:
            if timeout and (time.time() - start_time) > timeout:
                raise TimeoutError(f"Waiting for job {job_id} timed out after {timeout} seconds")
            time.sleep(interval)
        return self.get_result(job_id)

    def get_pipeline_names(self) -> list[str]:
        raise NotImplementedError()

    def get_param_schema(self, pipeline_name: str) -> dict | None:
        """Get workflow schema JSON dict (JSON Schema standard) or None if not available."""
        return None

    def get_failed_message(self, job_id):
        return f"Job {job_id} has failed."



import os
import shutil
import tempfile
from dataclasses import is_dataclass, fields

from sqlalchemy.orm import make_transient

from ovo import db, storage, config
from ovo.core.configuration import ConfigProps, load_config, save_default_config
from ovo.core.database import SqlDBEngine, DBEngine, DataclassType
from ovo.core.database.descriptors import ALL_DESCRIPTORS_BY_KEY
from ovo.core.database.models import (
    Project,
    Round,
    Pool,
    Design,
    DesignJob,
    DescriptorJob,
    DescriptorValue,
    FileDescriptor,
    ProjectArtifact,
)
from ovo.core.storage import Storage
from ovo.core.utils.formatting import safe_filename

# Configuration for import/export operations (Project handled separately)
IMPORT_EXPORT_CONFIGS = [
    {"model_class": Round, "filter": lambda o: dict(project_id__in=[p.id for p in o["project"]])},
    {"model_class": Pool, "filter": lambda o: dict(round_id__in=[r.id for r in o["round"]])},
    {"model_class": Design, "filter": lambda o: dict(pool_id__in=[p.id for p in o["pool"]])},
    {
        "model_class": DesignJob,
        "filter": lambda o: dict(id__in=[p.design_job_id for p in o["pool"] if p.design_job_id is not None]),
    },
    {"model_class": DescriptorJob, "filter": lambda o: dict(project_id__in=[p.id for p in o["project"]])},
    {
        "model_class": DescriptorValue,
        "filter": lambda o: dict(descriptor_job_id__in=[j.id for j in o["descriptor_job"]]),
    },
    {"model_class": ProjectArtifact, "filter": lambda o: dict(project_id__in=[p.id for p in o["project"]])},
]


def find_storage_path_fields(cls, prefix: str = "") -> list[str]:
    result = []
    if not is_dataclass(cls):
        return result

    for f in fields(cls):
        field_name = f.name
        path = f"{prefix}.{field_name}" if prefix else field_name
        metadata = f.metadata or {}

        # Check if it's a nested dataclass
        if is_dataclass(f.default) or (callable(f.default_factory) and is_dataclass(f.default_factory())):
            sub_cls = f.default if is_dataclass(f.default) else f.default_factory()
            result.extend(find_storage_path_fields(type(sub_cls), prefix=path))

        # Add to result only if metadata has storage_path=True
        elif metadata.get("storage_path") is True:
            result.append(path)

    return result


def export_import_project(
    source_db: DBEngine,
    source_storage: Storage,
    dest_db: DBEngine,
    dest_dir: str,
    project_id: str = None,
    check_conflicts=True,
    count_only=False,
    accepted_only=False,
) -> dict:
    """
    Unified function to copy a project from source database/storage to destination database/storage.

    Args:
        source_db: Database engine to read from
        dest_db: Database engine to write to
        project_id: ID of the project to copy (if None, copy all projects)
        source_storage: Storage instance to read files from
        dest_dir: Destination storage directory
        check_conflicts: Whether to check for ID conflicts in destination database
        count_only: If True, only count entities without copying them
        accepted_only: Only include accepted Designs

    Returns:
        dict: Summary with counts of copied entities
    """

    # Get the project from source database
    projects = source_db.select(Project, **(dict(id=project_id) if project_id else dict()))

    # Check for Project ID conflicts if requested (but not when only counting)
    if check_conflicts and not count_only and dest_db.count(Project, id__in=[p.id for p in projects]):
        existing_projects = dest_db.select(Project, id__in=[p.id for p in projects])
        names = ", ".join(f"{project.name} ({project.id})" for project in existing_projects)
        raise ValueError(f"Project already exists in the database: {names}")

    # Collect all storage file paths
    storage_paths = []
    all_objects = {"project": projects}
    counts = {
        "project": len(all_objects["project"]),
    }

    # Build all_objects using the configuration-driven approach
    for entity_config in IMPORT_EXPORT_CONFIGS:
        model_class = entity_config["model_class"]
        table_name = model_class.__tablename__

        filter_dict = entity_config["filter"](all_objects)

        if accepted_only:
            # Filter for accepted designs and their descriptors
            if model_class is Design:
                filter_dict["accepted"] = True
            if model_class is DescriptorValue:
                filter_dict["design_id__in"] = [d.id for d in all_objects["design"]]

        # Use the filter function to get filter dict and fetch objects
        print(f"Fetching {model_class.__name__} from DB for export")
        all_objects[table_name] = source_db.select(model_class, **filter_dict)
        print(f" Fetched {len(all_objects[table_name]):,} {model_class.__name__} objects")

        # Update counts
        counts[table_name] = len(all_objects[table_name])

        # Check for conflicts if requested and model has an id field and we have objects (but not when only counting)
        if check_conflicts and not count_only and hasattr(model_class, "id") and all_objects[table_name]:
            print("- Checking for conflicts...")
            obj_ids = [obj.id for obj in all_objects[table_name]]
            if dest_db.count(model_class, id__in=obj_ids):
                existing_ids = [obj.id for obj in dest_db.select(model_class, id__in=obj_ids)]
                raise ValueError(
                    f"{len(existing_ids):,} {model_class.__name__} objects already exist in the database: {existing_ids}"
                )

        # Collect storage files from objects
        for column in model_class.__table__.columns:
            # Inspect all columns marked with storage_path=True
            if column.info.get("storage_path"):
                for obj in all_objects[table_name]:
                    if file_path := getattr(obj, column.name, None):
                        storage_paths.append(file_path)
            # Inspect all fields marked with storage_path=True inside nested dataclass columns
            if isinstance(column.type, DataclassType):
                subclasses = set(
                    type(getattr(obj, column.name, None))
                    for obj in all_objects[table_name]
                    if getattr(obj, column.name, None)
                )
                for path_field in sorted(set(field for cls in subclasses for field in find_storage_path_fields(cls))):
                    for obj in all_objects[table_name]:
                        v = getattr(obj, column.name, None)
                        for field_name in path_field.split("."):
                            v = getattr(v, field_name, None)
                        if v is not None:
                            if isinstance(v, str):
                                storage_paths.append(v)
                            elif isinstance(v, list):
                                storage_paths.extend(v)
                            else:
                                raise ValueError(
                                    f"Unsupported storage path field type {type(v)} in {model_class.__name__}.{column.name}.{path_field}"
                                )

        if table_name == DescriptorValue.__tablename__:
            for obj in all_objects[table_name]:
                if ALL_DESCRIPTORS_BY_KEY.get(obj.descriptor_key) and isinstance(
                    ALL_DESCRIPTORS_BY_KEY[obj.descriptor_key], FileDescriptor
                ):
                    storage_paths.append(obj.value)

        if table_name == ProjectArtifact.__tablename__:
            for obj in all_objects[table_name]:
                if obj.artifact:
                    artifact_paths = obj.artifact.get_storage_paths()
                    storage_paths.extend(artifact_paths)

    print("All entities fetched from DB!")
    counts["storage_file"] = len(storage_paths)

    # If only counting, return counts
    if count_only:
        return counts

    # Copy storage files
    print(f"Copying {len(storage_paths):,} storage files...")
    source_storage.sync_files(storage_paths=storage_paths, local_destination_dir=dest_dir, preserve_subdirs=True)
    print("Storage copied successfully!")

    # Copy all objects to destination database
    print("Inserting entities to DB...")
    for project in projects:
        # handle project name conflicts
        while dest_db.count(Project, name=project.name, author=project.author):
            project.name = f"{project.name} (Copy)"

    for table_name, objects in all_objects.items():
        for obj in objects:
            make_transient(obj)
        dest_db.save_all(objects)
    print("Entities inserted successfully!")

    return counts


def export_project(project_id: str, output_zip_path: str = None, accepted_only: bool = False) -> str:
    """
    Export a single project and all associated data to a ZIP file containing
    ovo.db (SQLite database), storage subdirectory with all files, and config.yml.

    Args:
        project_id: ID of the project to export
        output_zip_path: Optional output zip path. Temporary file path will be used if not provided.
        accepted_only: Only export Designs that are accepted.

    Returns:
        str: Path to the exported ZIP file
    """

    # Get the project to verify it exists and get its name
    project = db.get(Project, project_id)

    # Create temporary directory for export
    with tempfile.TemporaryDirectory() as temp_root:
        temp_home = os.path.join(temp_root, safe_filename(project.name))
        os.makedirs(temp_home, exist_ok=True)

        # Add minimal config file
        save_default_config(temp_home, config_props=ConfigProps(read_only=True))

        # Create new SQLite database using SqlDBEngine
        export_db_path = os.path.join(temp_home, "ovo.db")
        export_db = SqlDBEngine(db_url=f"sqlite:///{export_db_path}")
        export_db.init()

        # Use unified function to copy project data
        counts = export_import_project(
            source_db=db,
            source_storage=storage,
            dest_db=export_db,
            dest_dir=os.path.join(temp_home, "storage"),
            project_id=project_id,
            accepted_only=accepted_only,
            check_conflicts=False,  # No need to check conflicts when exporting
        )

        if not project.public:
            print("NOTE: Project is private but will be exported as public!")
            export_db.save_value(Project, "public", True, id=project.id)

        # Add a README file
        readme_content = f"""
        OVO Project Export
    
        Project: {project.name}
    
        Contents:
        - ovo.db: SQLite database with all project data
        - storage/: Directory containing all project files with relative paths
        - config.yml: Minimal configuration file template
    
        To import this data:
        1. Use the Import & Export admin page in OVO
        2. Upload this ZIP file
        3. The system will restore the project data and files
        
        To preview this data:
        1. Unpack this ZIP file
        2. set OVO_HOME environment variable to the unpacked directory path
        3. run "ovo app"
    
        Export Summary:
        - Rounds: {counts["round"]}
        - Pools: {counts["pool"]}
        - Designs: {counts["design"]}
        - Storage files: {counts["storage_file"]}
        - Descriptor value: {counts["descriptor_value"]}
        - Project artifacts: {counts["project_artifact"]}
    """.lstrip()

        with open(os.path.join(temp_home, "README.txt"), "wt") as f:
            f.write(readme_content)

        # Create the final ZIP file
        print("Creating ZIP archive...")
        temp_zip_path = temp_root.removesuffix("/") + ".zip"
        shutil.make_archive(
            temp_zip_path.removesuffix(".zip"),  # archive path (without file extension)
            "zip",  # archive format
            temp_root,  # directory to archive
        )

    if output_zip_path:
        shutil.move(temp_zip_path, output_zip_path)
        return output_zip_path

    # Return path to the temporary ZIP file - should be deleted by the caller!
    return temp_zip_path


def import_project(home_dir: str, project_id: str = None, count_only=False) -> dict[str, int]:
    """
    Import project from extracted directory.

    Args:
        home_dir: Path to extracted directory containing ovo.db, storage files and config.yml
        project_id: ID of the project to import (if None, import all projects)

    Returns:
        counts: number of
    """

    # Load config from the extracted directory
    source_config = load_config(home_dir)

    # Use unified function to copy project data (no metadata transform)
    return export_import_project(
        source_db=SqlDBEngine(db_url=source_config.db.url),
        source_storage=Storage(storage_root=source_config.storage.path, aws=None),
        dest_db=db,
        dest_dir=config.storage.path,
        project_id=project_id,
        check_conflicts=True,
        count_only=count_only,
    )



import os

import streamlit as st

from ovo import storage, Pool
from ovo.app.components.custom_elements import confirm_download_button
from ovo.app.utils.cached_db import get_cached_design_jobs, get_cached_available_descriptors
from ovo.core.database import DesignWorkflow, WorkflowTypes, Base, FileDescriptor, DescriptorValue
from ovo.core.database.descriptors import ALL_DESCRIPTORS
from ovo.core.logic.descriptor_logic import export_design_descriptors_excel, get_wide_descriptor_table
from ovo.core.logic.design_logic import collect_storage_paths
from ovo.core.utils.formatting import get_hash_of_bytes


@st.fragment
def download_job_designs_component(
    design_ids: list[str], pools: list[Pool], key: str = "default", single_line: bool = True
):
    """
    Download design files for the specified workflow.
    Args:
        design_ids (list[str]): A list of design IDs to be displayed.
        pools (list[Pool]): A list of pools associated with the designs.
        key (str): Streamlit component key prefix
        single_line: Whether to display the download options in a single line
    """

    # label -> (Model, field_name)
    # for example: {"pdb files": (Design, "structure_path")}
    download_fields = {}
    if any(not pool.design_job_id for pool in pools):
        # Some pools don't have a design job (e.g. uploaded designs), so we need to include the basic fields
        download_fields.update(DesignWorkflow.get_download_fields())
    design_jobs = get_cached_design_jobs([pool.design_job_id for pool in pools if pool.design_job_id])
    for design_job in design_jobs:
        WorkflowType = (
            WorkflowTypes.get(design_job.workflow.name) if design_job and design_job.workflow else DesignWorkflow
        )
        for label, v in WorkflowType.get_download_fields().items():
            if label in download_fields:
                assert download_fields[label] == v, (
                    f"Conflicting download field for key '{label}': {download_fields[label]} != {v}"
                )

        download_fields.update(WorkflowType.get_download_fields())
    # Add available file descriptor paths
    available_descriptors = get_cached_available_descriptors(design_ids)
    file_descriptors = [d for d in ALL_DESCRIPTORS if isinstance(d, FileDescriptor) and d.key in available_descriptors]
    for file_descriptor in file_descriptors:
        download_fields[file_descriptor.name] = (DescriptorValue, file_descriptor.key)

    if single_line:
        first, second, third, _ = st.columns([1, 1, 1, 1])
    else:
        first = st.container()
        second = st.container()
        third = st.container()
    # for example ovo_xyv_avg_123_designs
    if len(design_ids) <= 3:
        filename_prefix = "_".join(design_ids)
    elif len(pools) < 10:
        filename_prefix = "ovo_" + "_".join(p.id for p in pools)
    else:
        filename_prefix = f"ovo_{len(pools)}_pools"
    filename = filename_prefix + f"_{len(design_ids)}_designs"
    key = key + "_" + get_hash_of_bytes(",".join(design_ids).encode())

    # Descriptor table
    with first:
        download_descriptor_table(filename, design_ids, key=key, width="stretch")

    # All files in one zip
    with second:
        download_design_files(
            label="design files",
            download_fields=download_fields,
            filename=filename,
            design_ids=design_ids,
            key=key,
        )

    # Individual file types in separate zips
    if len(download_fields) > 1:
        with third:
            with st.popover("Download by type", width="stretch"):
                for label, (Model, field) in download_fields.items():
                    download_design_files(
                        label=label,
                        download_fields={label: (Model, field)},
                        filename=f"{filename}_{label.lower().replace(' ', '_')}",
                        design_ids=design_ids,
                        key=key,
                    )


@st.fragment
def download_descriptor_table(filename, design_ids, descriptor_keys=None, key="default", width="content"):
    if st.button(
        "Download descriptor table" if len(design_ids) > 1 else "Download descriptors",
        key=f"prepare_descriptors_{key}",
        width=width,
    ):
        with st.spinner("Preparing descriptor table..."):
            # Get raw dataframe with single header, columns named with descriptor keys ("pipeline|tool_key|descriptor")
            df = get_wide_descriptor_table(
                design_ids=design_ids, descriptor_keys=descriptor_keys, nested=False, human_readable=False
            )
            excel_bytes = export_design_descriptors_excel(df)
        confirm_download_button(
            data=excel_bytes.getvalue(),
            file_name=f"{filename}.xlsx",
            mime="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
            key=f"download_descriptors_{key}",
        )


@st.fragment
def download_design_files(
    label: str, download_fields: dict[str, tuple[Base, str]], filename: str, design_ids: list, key="default"
):
    if st.button(f"Download {label}", key=f"prepare_{filename}_{key}", width="stretch"):
        with st.spinner("Loading files..."):
            # Get list of storage paths from Design or DesignWorkflow objects
            storage_paths = collect_storage_paths(download_fields, design_ids)

        if not storage_paths:
            st.error(f"No files found")
            return
        elif len(storage_paths) == 1:
            # Single file, download directly
            filename = os.path.basename(storage_paths[0])
            with st.spinner(f"Downloading file..."):
                data = storage.read_file_bytes(storage_paths[0])
            st.write(f"Prepared file *{filename}*")
            confirm_download_button(data=data, file_name=filename, key=f"download_{filename}_{key}")
        else:
            # Multiple files, prepare zip
            with st.spinner(f"Preparing zip file with {len(storage_paths):,} results..."):
                data = storage.create_zip({filename: storage_paths})

            st.write(
                f"Prepared zip with {len(storage_paths):,} files for {len(design_ids):,} {'design' if len(design_ids) == 1 else 'designs'}."
            )
            confirm_download_button(
                data=data, file_name=filename + ".zip", mime="application/zip", key=f"download_{filename}_{key}"
            )



import os
import argparse
import glob

from esm import pretrained
import torch
import esm.inverse_folding
from esm.inverse_folding.util import CoordBatchConverter

import pandas as pd


def load_model(models_path):
    model = "esm_if1_gvp4_t16_142M_UR50.pt"
    model_path = os.path.join(models_path, model)
    model, alphabet = pretrained.load_model_and_alphabet(model_path)
    return (model, alphabet)


def get_sequence_aa_probs(model, alphabet, coords, seq, softmax=True):
    device = next(model.parameters()).device
    batch_converter = CoordBatchConverter(alphabet)
    batch = [(coords, None, seq)]
    coords, confidence, strs, tokens, padding_mask = batch_converter(batch, device=device)

    prev_output_tokens = tokens[:, :-1].to(device)
    target = tokens[:, 1:]
    target_padding_mask = target == alphabet.padding_idx
    logits, _ = model.forward(coords, padding_mask, confidence, prev_output_tokens)
    if softmax:
        probs = torch.softmax(logits, dim=1).cpu().detach().numpy()[0].T
    else:
        probs = logits.cpu().detach().numpy()[0].T
    return pd.DataFrame(probs, columns=alphabet.all_toks)[list("-ACDEFGHIKLMNPQRSTVWXY")]


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_path", type=str)
    parser.add_argument("output_csv", type=str)
    parser.add_argument("-c", "--chains", type=str, default="A", help="Chain to extract the sequence from")
    parser.add_argument("--esm_models_dir", type=str, help="Path to the ESM models directory")
    options = parser.parse_args()

    paths = sorted(glob.glob(os.path.join(options.input_path, "*.pdb")))
    if not paths:
        raise ValueError(f"No pdb files found in: {options.input_path}")

    print(f"Processing {len(paths):,} structures")

    models_path = options.esm_models_dir

    model, alphabet = load_model(models_path)
    print("Model loaded")
    model = model.eval()
    if torch.cuda.is_available():
        model = model.cuda()
        print("Transferred model to GPU")

    chains = [chain.strip() for chain in options.chains.split(",")]

    results = []
    for i, path in enumerate(paths):
        print(f"Processing {i + 1}/{len(paths)}: {path}")
        basename = os.path.basename(path).removesuffix(".pdb")
        structure = esm.inverse_folding.util.load_structure(path)
        coords, native_seqs = esm.inverse_folding.multichain_util.extract_coords_from_complex(structure)

        chain_softmax = []
        for chain_id in chains:
            if not (chain_id in coords and chain_id in native_seqs):
                raise ValueError(
                    f"Chain {chain_id} not found in structure: {path}, available chains: {list(coords.keys())}"
                )
            print("Processing chain:", chain_id)
            native_seq = native_seqs[chain_id]
            all_coords = esm.inverse_folding.multichain_util._concatenate_coords(coords, chain_id)
            logits = get_sequence_aa_probs(model, alphabet, all_coords, native_seq, softmax=False)
            softmax = pd.DataFrame(torch.softmax(torch.from_numpy(logits.values), dim=1), columns=logits.columns)
            softmax["native_aa"], softmax["chain_id"] = list(native_seq), chain_id
            chain_softmax.append(softmax)

        df_softmax = pd.concat(chain_softmax)
        native_seq_sum_softmax = sum([row[row["native_aa"]] for (i, row) in df_softmax.iterrows()])
        # Compute average softmax probability per structure given total sequence length (concatenated sequences of processed chain)
        native_seq_avg_softmax = native_seq_sum_softmax / len(df_softmax["native_aa"])
        processed_chains = ",".join(df_softmax["chain_id"].unique())
        results.append((basename, processed_chains, native_seq_avg_softmax))

    result = pd.DataFrame(results, columns=["id", "chain", "native_seq_avg_softmax"])
    result.to_csv(options.output_csv, index=False)



from pathlib import Path
import streamlit as st

from ovo import config
from ovo.app.utils.page_init import initialize_page

# We have to initialize at every page because of 2 reasons:
# - importing the styles (for the OVO logo etc., contigs styling, setting the favicon...)
# - to make sure that the st.session_state contains all variables used throughout the app

initialize_page("OVO, where new proteins hatch!")

st.title("Welcome to OVO" + (f", {st.user.given_name}!" if st.user.get("given_name") else "!"))

with st.container(width=800):
    if config.templates.welcome:
        st.markdown(Path(config.templates.welcome).read_text(), unsafe_allow_html=True)

    st.markdown(f"""
    
### Getting started

Please select a *design* workflow in the left panel.
Each step-by-step submission form is customized to help you select all workflow parameters for a specific design use case. 
Submitting the workflow will create a *job*. 
Each job produces a *pool* of designs. You can also use OVO to upload a custom pool structures or sequences for analysis.
        
The project dropdown in the left sidebar allows you to organize your jobs and designs into *projects*.
Jobs in each project can be organized into *rounds*.

The ⏳ *Jobs* page shows all running and finished jobs neatly together with their workflow parameters. 
In each *job result* page, you can explore all generated designs and their descriptors such as AlphaFold2 RMSD and Rosetta ddG. 
You will be able to adjust *acceptance thresholds* to find the most promising designs we call *accepted designs*.

The 🐣 *Designs* page focuses further on the *accepted designs*. The 🔵 *Explorer* view enables plotting
design descriptors and visualizing their structures. The 🔎 *ProteinQC* view provides additional computational evaluation
informing you about the quality of the designs using additional descriptors focused on 
physicochemical properties, sequence composition, and fitness from protein language models.

### Learning resources

- [User Guide](https://ovo.dichlab.org/docs/user_guide/)
- [Ovo Overview Video](https://github.com/user-attachments/assets/7b339fa6-c6de-467d-90d0-5cd15f83c498)
""")

if config.templates.welcome_appendix:
    st.markdown(Path(config.templates.welcome_appendix).read_text(), unsafe_allow_html=True)
else:
    st.markdown("""
    ---
    Report bugs and feature requests at [github.com/MSDLLCpapers/ovo/issues](https://github.com/MSDLLCpapers/ovo/issues)            
    """)



import dataclasses
import json
from typing import Literal


@dataclasses.dataclass
class ContigSegment:
    """Class for storing data about individual contig regions."""

    value: str
    length: int
    type: str
    input_res_start: int
    input_res_end: int
    input_res_chain: str
    out_res_start: int
    out_res_end: int
    out_res_chain: str
    color: str | None = None


class EnhancedJSONEncoder(json.JSONEncoder):
    """A JSON encoder for enabling the export of dataclasses."""

    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)  # type: ignore
        return super().default(o)


@dataclasses.dataclass
class ChainVisualization:
    """Class for storing data about individual chains.

    Parameters
    ----------
    chain_id : str
        The chain ID.
    color : Literal["uniform", "chain-id", "hydrophobicity", "plddt", "molecule-type", "secondary-structure", "residue-name", "residue-charge"], optional
        The color scheme to use, by default "uniform".
    color_params : dict, optional
        Color parameters.
        In the case of "uniform", the `color_params` field should be for example {"value": "0x00ff00"}.
        In the case of "chain-id", the `color_params` field should be for example {"palette": "pastel-1"}.
    representation_type : Literal["cartoon", "cartoon+ball-and-stick", "molecular-surface", "gaussian-surface", "ball-and-stick"], optional
        The representation type to use, by default "cartoon".
    residues : list[int], optional
        Create a representation just for the specified residues, by default None (all residues).
    label : str, optional
        The label to show when hovering over the visualization, by default None.
    """

    chain_id: str
    color: Literal[
        "uniform",
        "chain-id",
        "hydrophobicity",
        "plddt",
        "molecule-type",
        "secondary-structure",
        "residue-name",
        "residue-charge",
    ] = "uniform"
    color_params: dict | None = None
    representation_type: Literal[
        "cartoon", "cartoon+ball-and-stick", "molecular-surface", "gaussian-surface", "ball-and-stick"
    ] = "cartoon"
    residues: list[int] | None = None
    label: str | None = None


@dataclasses.dataclass
class StructureVisualization:
    """Class for storing data about structures shown in the Mol* component.

    Parameters
    ----------
    pdb : str
        A PDB/mmCIF content of the structure to visualize. Can also be a URL to a PDB/mmCIF file.
    contigs : list[ContigSegment], optional
        A list of contig segments to show (e.g. on the top of the viewer), by default None.
    color : Literal["uniform", "chain-id", "hydrophobicity", "plddt", "molecule-type", "secondary-structure", "residue-name", "residue-charge"], optional
        The color scheme to use, by default "uniform".
    color_params : dict, optional
        Color parameters.
        In the case of "uniform", the `color_params` field should be for example {"value": "0x00ff00"}.
        In the case of "chain-id", the `color_params` field should be for example {"palette": "pastel-1"}.
    representation_type : Literal["cartoon", "cartoon+ball-and-stick", "molecular-surface", "gaussian-surface", "ball-and-stick"], optional
        The representation type to use, by default "cartoon".
    highlighted_selections : list[str], optional
        A list of selections to select and highlight in green color, by default None.
    chains : list[ChainVisualization], optional
        A list of separate chain visualizations, useful for individual coloring and representation, by default None. Available only for PDB files in a string format.
    """

    pdb: str
    contigs: str | list[ContigSegment] | None = None
    color: Literal[
        "uniform",
        "chain-id",
        "hydrophobicity",
        "plddt",
        "molecule-type",
        "secondary-structure",
        "residue-name",
        "residue-charge",
    ] = "uniform"
    color_params: dict | None = None
    representation_type: (
        Literal["cartoon", "cartoon+ball-and-stick", "molecular-surface", "gaussian-surface", "ball-and-stick"] | None
    ) = "cartoon"
    highlighted_selections: list[str] | None = None
    chains: list[ChainVisualization] | None = None

    def __post_init__(self):
        if self.chains:
            if isinstance(self.chains, ChainVisualization):
                self.chains = [self.chains]
            elif not isinstance(self.chains, list):
                raise ValueError(
                    f"Invalid type for chains, expected ChainVisualization or list of ChainVisualization, got: {type(self.chains).__name__}"
                )
        if self.contigs:
            from ovo.app.components.molstar_custom_component import ContigsParser

            parser = ContigsParser()
            if isinstance(self.contigs, str):
                parsed_contigs = parser.parse_contigs_str(self.contigs)
            else:
                assert isinstance(self.contigs, list), (
                    f"Invalid type for contigs, expected str or list, got: {type(self.contigs).__name__}"
                )
                parsed_contigs = []
                for contig_or_segment in self.contigs:
                    if isinstance(contig_or_segment, str):
                        parsed_contigs.extend(parser.parse_contigs_str(contig_or_segment))
                    elif isinstance(contig_or_segment, ContigSegment):
                        parsed_contigs.append(contig_or_segment)
                    else:
                        raise ValueError(
                            f"Invalid type for contig segment, expected str or ContigSegment, got: {type(contig_or_segment).__name__}"
                        )
            self.contigs = parsed_contigs

    def to_dict(self):
        return dataclasses.asdict(self)



import streamlit as st
import numpy as np
from dataclasses import dataclass
import pandas as pd
from typing import Dict

from ovo.core.database.models import Descriptor

from ovo.app.components.descriptor_scatterplot import format_descriptor_name
from ovo.app.utils.protein_qc_plots import (
    format_threshold_value,
    descriptor_histograms,
    source_selectbox,
)


@dataclass
class HistogramSettings:
    descriptor: Descriptor | None = None
    source: str | None = None

    @classmethod
    def from_query_params(cls, descriptors_by_key: dict[str, Descriptor], key_prefix: str):
        return cls(
            descriptor=descriptors_by_key[st.query_params[f"{key_prefix}_descriptor"]]
            if f"{key_prefix}_descriptor" in st.query_params and f"{key_prefix}_descriptor" in descriptors_by_key
            else None,
            source=st.query_params[f"{key_prefix}_source"] if f"{key_prefix}_source" in st.query_params else None,
        )

    def update_query_params(self, key_prefix: str):
        if self.descriptor:
            st.query_params[f"{key_prefix}_descriptor"] = self.descriptor.key
        else:
            st.query_params.pop(f"{key_prefix}_descriptor", None)
        if self.source:
            st.query_params[f"{key_prefix}_source"] = self.source
        else:
            st.query_params.pop(f"{key_prefix}_source", None)


@st.fragment()
def descriptor_explorer(
    descriptors_df: pd.DataFrame, descriptors_by_key: Dict[str, Descriptor], single_design: bool = False
):
    if descriptors_df.empty:
        st.warning("No designs found")
        return

    if not descriptors_by_key:
        st.warning("No descriptors found for the selected designs.")
        return

    key_prefix = "design_expl" if single_design else "descriptor_expl"

    settings = HistogramSettings.from_query_params(descriptors_by_key, key_prefix)

    descriptor_col, source_col, _ = st.columns([2, 1, 1])

    # Select the descriptor to plot
    descriptor_options = list(descriptors_by_key.keys())
    with descriptor_col:
        descriptor = st.selectbox(
            "Descriptor",
            format_func=lambda descriptor_key: format_descriptor_name(descriptors_by_key[descriptor_key]),
            index=descriptor_options.index(settings.descriptor.key) if settings.descriptor is not None else 0,
            key=f"{key_prefix}_descriptor",
            options=descriptor_options,
            # force_change=preset_changed,
        )
        settings.descriptor = descriptors_by_key.get(descriptor)

    with source_col:
        settings.source = source_selectbox(settings.source, key_prefix=key_prefix)

    if not settings.descriptor:
        st.warning("Please select a descriptor to plot.")
        return

    settings.update_query_params(key_prefix)
    descriptor_values = descriptors_df[(settings.descriptor.tool, settings.descriptor.name)]
    nonnull_descriptor_values = descriptor_values.dropna().tolist()

    # Write the descriptor description and thresholds
    st.write(f"*{settings.descriptor.description}*")

    if single_design:
        avg_col, warning_col, error_col = st.columns(3)
    else:
        avg_col, median_col, warning_col, error_col = st.columns(4)

    # If there are missing values, show warning message that the distribution is only for the non-missing values
    if len(descriptor_values) > len(nonnull_descriptor_values):
        perc_missing = (1 - len(nonnull_descriptor_values) / len(descriptor_values)) * 100
        st.warning(
            f"This descriptor is missing for **{perc_missing:.2f}%** of designs. The distribution was calculated from designs with non-missing values. You can submit **full ProteinQC** to calculate the descriptor for all designs.",
            icon=":material/warning:",
        )
    with avg_col:
        avg_value = np.mean(descriptor_values)
        avg_value = format_threshold_value(settings.descriptor.name, avg_value)
        label = (
            "" if single_design else "Average "
        ) + settings.descriptor.name  # f'Average {settings.descriptor.name}'
        st.metric(label, avg_value)
    if not single_design:
        with median_col:
            median_value = np.median(descriptor_values)
            median_value = format_threshold_value(settings.descriptor.name, median_value)
            st.metric(f"Median {settings.descriptor.name}", median_value)
    with warning_col:
        warning_value = format_threshold_value(settings.descriptor.name, settings.descriptor.warning_value)
        st.metric("Warning threshold", warning_value)
    with error_col:
        error_value = format_threshold_value(settings.descriptor.name, settings.descriptor.error_value)
        st.metric("Error threshold", error_value)

    fig = descriptor_histograms(
        descriptor=settings.descriptor,
        histogram_source=settings.source,
        values=descriptor_values,
        precomputed_histogram_height=200,
        pool_histogram_height=200,
    )

    st.altair_chart(
        fig,
        width="stretch" if single_design else "content",
        key=f"descriptor_histogram_{key_prefix}",
    )



from abc import ABC
from typing import List, Callable, Optional

from ovo.core.database import DescriptorJob
from ovo.core.database.descriptors_refolding import REFOLDING_TESTS_BY_TYPE
from ovo.core.database.models import DescriptorWorkflow, WorkflowTypes, Base
from dataclasses import dataclass, field


class RefoldingSupportedDesignWorkflow(ABC):
    """Abstract class that defines an interface for workflows supported by the RefoldingWorkflow workflow"""

    def get_refolding_native_pdb_path(self, contig_index: int) -> Optional[str]:
        raise NotImplementedError()

    def get_refolding_design_paths(self, design_ids: list[str]) -> dict[str, str]:
        raise NotImplementedError()

    def get_refolding_design_type(self) -> str:
        raise NotImplementedError()


@WorkflowTypes.register("Refolding workflow")
@dataclass
class RefoldingWorkflow(DescriptorWorkflow):
    chains: List[str] = field(default_factory=lambda: ["FULL"])
    tests: List[str] = None
    design_type: str = None
    native_pdb_path: str = None

    def get_pipeline_name(self) -> str:
        return "ovo.refolding"

    def prepare_params(self, workdir: str) -> dict:
        from ovo.core.logic.descriptor_logic import prepare_refolding_params

        return prepare_refolding_params(self, workdir=workdir)

    def process_results(self, job: DescriptorJob, callback: Callable = None) -> list[Base]:
        from ovo.core.logic.descriptor_logic import read_descriptor_file_values

        descriptor_values = read_descriptor_file_values(
            descriptor_job=job,
            # descriptor key prefix (pipeline|tool_key) -> filename to parse
            filenames={f"refolding|{test}": test for test in self.tests},
            # mapping from design.id to ID column in produced file
            design_id_mapping={design_id: design_id for design_id in self.design_ids},
        )
        return descriptor_values + [job]

    def validate(self):
        super().validate()
        if not self.design_type:
            raise ValueError("Design type not specified")
        if not self.tests:
            raise ValueError("No tests specified")
        for test in self.tests:
            # TODO validate test name
            assert test in REFOLDING_TESTS_BY_TYPE[self.design_type], (
                f"Invalid test for design type {self.design_type}: {test}"
            )



import os
import re

import streamlit as st
from ovo.app.components.contigs_organizer import contigs_organizer
from ovo.app.components.molstar_custom_component import (
    molstar_custom_component,
    StructureVisualization,
    ContigsParser,
    ChainVisualization,
)

from ovo import storage
from ovo.core.database.models_rfdiffusion import (
    RFdiffusionScaffoldDesignWorkflow,
    RFdiffusionWorkflow,
    RFdiffusionBinderDesignWorkflow,
)
from ovo.app.components.molstar_custom_component.dataclasses import ContigSegment
from ovo.app.components.trim_components import check_hotspots
from ovo.core.utils.pdb import add_glycan_to_pdb, filter_pdb_str
from ovo.core.utils.residue_selection import (
    from_segments_to_hotspots,
    from_hotspots_to_segments,
    from_contig_to_residues,
    from_residues_to_segments,
)


def parameters_binder_preview_component(workflow: RFdiffusionBinderDesignWorkflow):
    if not workflow.target_chain:
        st.error("Please provide a target chain in the previous step.")
        return

    st.write(f"Target chain: {workflow.target_chain}")
    st.write(f"Start and end residue: {workflow.start_res_trimmed_chain} - {workflow.end_res_trimmed_chain}")

    binder_length_col, hotspots_col, _ = st.columns([1, 1, 3])

    with binder_length_col:
        binder_length = st.text_input(
            "Binder length (min-max)",
            placeholder="For example 20-40",
            value=workflow.binder_length,
            key="binder_length",
        )
        if binder_length != workflow.binder_length:
            # Reset the contig to force regeneration below
            workflow.rfdiffusion_params.contig = None
            workflow.binder_length = binder_length

    with hotspots_col:
        workflow.rfdiffusion_params.hotspots = st.text_input(
            "Hotspot residues (optional)",
            placeholder="For example A123,A124,A131",
            value=workflow.rfdiffusion_params.hotspots,
            key="hotspots",
        )
        if workflow.rfdiffusion_params.hotspots:
            if not all(
                re.fullmatch("[A-Z][0-9]+", hotspot) for hotspot in workflow.rfdiffusion_params.hotspots.split(",")
            ):
                st.error("Invalid hotspots format, expected 'A123,A124,A131'")
                return

    # Check if the selection has changed since the last preview
    hotspots_from_first_step = from_segments_to_hotspots(workflow.selected_segments)
    if workflow.rfdiffusion_params.hotspots != hotspots_from_first_step:
        # Reset the preview_job_id to prevent the visualization of the previous preview
        workflow.preview_job_id = None
        # Update the selected segments based on the current hotspots
        workflow.selected_segments = from_hotspots_to_segments(workflow.rfdiffusion_params.hotspots)

    # Exit early if we don't have all fields required for contig
    if not workflow.target_chain or not workflow.binder_length:
        workflow.rfdiffusion_params.contig = None
        return

    # Exit early if binder length is invalid
    if not re.fullmatch(r"^\d+(-\d+)?$", binder_length):
        st.error("Binder length should be in this format: '20' or '20-40'")
        return

    # Update contig if not set yet, or if reset due to changing trim region or binder length
    if not workflow.rfdiffusion_params.contig:
        # Create contig definition for target, take trimming into account
        target_contig = from_residues_to_segments(
            workflow.target_chain,
            from_contig_to_residues(workflow.chains.get(workflow.target_chain)),
            workflow.start_res_trimmed_chain,
            workflow.end_res_trimmed_chain,
        )
        workflow.rfdiffusion_params.contig = "/".join(target_contig) + f"/0 {workflow.binder_length}"

    if workflow.rfdiffusion_params.hotspots:
        check_hotspots(workflow.rfdiffusion_params.contig, workflow.rfdiffusion_params.hotspots)


@st.fragment
def scaffold_contig_preview(pdb_input_string, parsed_contig: list[ContigSegment]):
    """Preview the fixed segments in the input structure.

    :param pdb_input_string: The input PDB string.
    :param parsed_contig: Contig segment objects
    """

    fixed_segments = [seg for seg in parsed_contig if seg.type == "fixed"]

    try:
        pdb_fixed_string = filter_pdb_str(
            pdb_input_string, segments=[seg.value for seg in fixed_segments], add_ter=True
        )

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=pdb_fixed_string,
                    contigs=fixed_segments,
                    representation_type="cartoon+ball-and-stick",
                )
            ],
            key="fixed_segments_preview",
            height=400,
        )
    except Exception:
        st.error("Invalid contig provided.")


@st.fragment
def contigs_organizer_fragment(page_key: str, pdb_input_string: str):
    workflow: RFdiffusionScaffoldDesignWorkflow = st.session_state.workflows[page_key]

    contig = update_contig_based_on_selected_segments(
        old_contig=workflow.get_contig(), selected_segments=workflow.get_selected_segments()
    )

    contig_input_key = "input_contig"
    if contig_input_key not in st.session_state:
        st.session_state[contig_input_key] = contig

    if st.session_state.get("contig_updated_from_component"):
        # force update input field when contig was changed from the custom component
        # we need to use a separate session key since streamlit only allows setting the key before calling the component
        st.session_state["contig_updated_from_component"] = False
        st.session_state[contig_input_key] = contig

    new_contig = st.text_input(
        "Contig",
        placeholder="A123-456/10/A567-890/...",
        key=contig_input_key,
    ).strip()

    if new_contig != workflow.get_contig():
        workflow.set_selected_segments([s for s in new_contig.split("/") if s and s[0].isalpha()])
        workflow.set_contig(new_contig)

    parser = ContigsParser()
    parsed_contig = parser.parse_contigs_str(
        workflow.get_contig(),
        # TODO
        # include_generated=True
        # We could try including generated segments to keep their color in contig organizer
        # The problem is that the color is assigned by segment value,
        # so all segments of same length will get the same color
        # Another problem is that we don't store which fixed residues are connected by each generated segment,
        # So without that information we can't visualize it as a link in the structure viewer.
    )

    left, right = st.columns([1.2, 1])

    with left:
        st.caption("Contig editor")

        if workflow.get_contig() and " " in workflow.get_contig():
            missing_chain_break = any(
                not inner_sub_contig.endswith("/0") for inner_sub_contig in workflow.get_contig().split(" ")[:-1]
            )
            if missing_chain_break:
                st.error(
                    "When contig contains multiple chains (separated by space), each chain should end with '/0' (chain break)."
                )
            # TODO elaborate on the limitations
            st.warning("Support for multi-chain contigs in scaffold design is experimental, proceed with caution.")
            return

        co = contigs_organizer(
            contigs=workflow.get_contig(),
            pdb=pdb_input_string,
            colors={s.value: s.color for s in parsed_contig},
            key="contigs_organizer",
        )

        if co:
            previous_contig = co.get("previous", {}).get("contig")
            # When passing a new contig into the component, it can still return its old state here,
            # until it gets updated to the new state (remember this happens asynchronously in the browser)
            # We only accept the change it if the "previous" state matches our current state
            if previous_contig == workflow.get_contig():
                # Update contig and selected segments based on response from component
                new_contig = co.get("current", {}).get("contig")
                if new_contig and workflow.get_contig() != new_contig:
                    workflow.set_selected_segments([s for s in new_contig.split("/") if s and s[0].isalpha()])
                    workflow.set_contig(new_contig)
                    st.session_state["contig_updated_from_component"] = True
                    st.rerun()

        if not workflow.preview_job_id and workflow.get_contig():
            try:
                all_segments = parser.parse_contigs_str(workflow.get_contig(), include_generated=True)
                num_generated_segments = sum(s.type == "generated" for s in all_segments)
                num_fixed_segments = sum(s.type == "fixed" for s in all_segments)
                if num_generated_segments and num_generated_segments >= num_fixed_segments - 1:
                    st.write("Contig looks ready. Generate a preview below :material/arrow_cool_down:")
            except Exception:
                # Silently ignore parsing contig, since it might not be ready yet
                pass

    with right:
        st.caption("Fixed segments")
        scaffold_contig_preview(pdb_input_string, parsed_contig)


def update_contig_based_on_selected_segments(old_contig: str, selected_segments: list[str]):
    contig = old_contig
    if old_contig is None:
        if selected_segments is None:
            # Generate new empty contig
            st.warning("No segments selected, please select them in the previous step or fill in the contig manually")
            contig = ""
        else:
            # Generate first contig template
            contig = "/".join(selected_segments or [])
    else:
        if " " in old_contig.strip():
            # do not try updating the contig if it contains a space (might be a mistake from the user, let them fix it below)
            # do not exit here - more errors will be shown below
            pass
        else:
            # Get all contig segments (including generated: A10-20/10/A30-40 -> ['A10-20', '10', 'A30-40'])
            contig_segments = [s for s in old_contig.strip().split("/") if s]
            # Get fixed selected segments from contig, for example [(0, 'A10-20'), (2, 'A30-40')]
            fixed_contig_segments_with_index = [(i, s) for i, s in enumerate(contig_segments) if s[0].isalpha()]
            fixed_contig_segments = [s for i, s in fixed_contig_segments_with_index]
            if fixed_contig_segments != selected_segments:
                # Regenerate contig if user changed the selection in previous step
                if len(fixed_contig_segments) == len(selected_segments) and any(
                    a == b for a, b in zip(fixed_contig_segments, selected_segments)
                ):
                    # Same number of fixed segments, and some segments are unchanged
                    # (that check is done to give up when segments are reordered,
                    #  since molstar discards user's ordering and returns selection in PDB order,
                    #  so it's better to just regenerate the contig in that case)
                    # -> Preserve contig, keep lengths of all generated segments in between
                    changed_segments = []
                    for (i, old_segment), new_segment in zip(fixed_contig_segments_with_index, selected_segments):
                        if old_segment != new_segment:
                            changed_segments.append(f"{old_segment} -> {new_segment}")
                            # replace contig_segment to selected_segment
                            assert old_segment == contig_segments[i], (
                                f"Expected segment {old_segment} at index {i} in {contig_segments}"
                            )
                            contig_segments[i] = new_segment
                    contig = "/".join(contig_segments)
                    st.success(f"""
                    Contig was updated based on changed segments: {", ".join(changed_segments)}
                    
                    Previous contig: {old_contig}
                    
                    New contig: {contig}
                    """)
                else:
                    # TODO we could preserve at least the generated segment lengths in between segments that haven't changed
                    contig = "/".join(selected_segments or [])
                    if fixed_contig_segments:
                        st.warning(f"""
                        Number of selected segments changed, generating new contig template.
                        
                        Previous contig: {old_contig}
                        
                        New contig: {contig}
                        """)
    return contig


def visualize_rfdiffusion_preview(workflow: RFdiffusionWorkflow, output_dir: str):
    pdb_preview = os.path.join(
        output_dir, "rfdiffusion", "rfdiffusion_standardized_pdb", "rfdiffusion_0_standardized.pdb"
    )
    trb_preview = os.path.join(output_dir, "rfdiffusion", "rfdiffusion_trb", "rfdiffusion_0.trb")

    if not os.path.exists(pdb_preview):
        st.error(
            f"The .pdb preview file does not exist. Something went wrong during the RFdiffusion process: {pdb_preview}"
        )
        return

    if not os.path.exists(trb_preview):
        st.error(f"The .trb file does not exist. Something went wrong during the RFdiffusion process: {trb_preview}")
        return

    # Serialize contigs
    parser = ContigsParser()
    parsed_contigs_trb = parser.parse_contigs_trb(trb_preview)

    with open(pdb_preview) as f:
        pdb_preview_string = f.read()

    pdb_input_string = storage.read_file_str(workflow.get_input_pdb_path())

    # Parse the contigs again to make sure that no empty contigs are provided...
    # We create a new instance to ensure the coloring stays the same
    parser = ContigsParser()
    parsed_contigs_str = parser.parse_contigs_str(workflow.get_contig())

    if st.toggle(
        "Show glycosylation sites",
        key="show_glycosylation_sites_input",
        help="Glycans are grafted to all NXS and NXT motifs (except NGS and NGT) "
        "based on a glycan template from PDB 5ELI. Nonexistent bonds might be displayed.",
    ):
        pdb_input_string, _ = add_glycan_to_pdb(pdb_input_string)
        pdb_preview_string, glycosylated_residues = add_glycan_to_pdb(pdb_preview_string)

        if glycosylated_residues:
            st.warning(f"Glycosylated motif found at {','.join(glycosylated_residues)}")
        else:
            st.info("No glycosylation motif found.")

    hotspot_segments = from_hotspots_to_segments(workflow.get_hotspots()) or []

    # visualize preview
    visual_col1, visual_col2 = st.columns(2)
    with visual_col1:
        st.write("Input structure")
        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=pdb_input_string,
                    contigs=parsed_contigs_str,
                    highlighted_selections=hotspot_segments,
                    representation_type="cartoon+ball-and-stick",
                )
            ],
            key="inp_structure",
        )
    with visual_col2:
        st.write("RFdiffusion design preview")

        # TODO
        # Note this assumes that the target chain is B in the RFdiffusion output PDB,
        # and that the residue numbers are same as in the input PDB
        output_hotspot_segments = [s.replace(workflow.get_target_chain(), "B") for s in hotspot_segments or []]

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=pdb_preview_string,
                    # TODO do not change logic based on instance type,
                    #  add methods instead
                    contigs=parsed_contigs_trb if workflow.is_instance(RFdiffusionScaffoldDesignWorkflow) else None,
                    highlighted_selections=output_hotspot_segments,
                    chains=[
                        ChainVisualization(
                            chain_id="A",
                            color_params={"value": "0xde853c"},
                        )
                    ]
                    if workflow.is_instance(RFdiffusionBinderDesignWorkflow)
                    else None,
                )
            ],
            key="preview_structure",
            download_filename="rfdiffusion_preview",
        )



import pandas as pd
from typing import List, Collection

from ovo import db, config
from ovo.core.database.cache_clearing import clear_when_modified
from ovo.core.database.models import Project, DescriptorValue, Descriptor, Pool, Round, Design, DesignJob
import streamlit as st

from ovo.core.logic.descriptor_logic import get_available_descriptors
from ovo.core.logic.design_logic import get_design_jobs_table, get_pools_table


@clear_when_modified(Project)
@st.cache_data(ttl="1h")
def get_cached_project_ids_and_names(username: str, extra_project_ids: Collection[str] = None) -> dict[str, str]:
    """Return dict of project id -> name"""
    if username in config.auth.admin_users:
        projects = db.select(Project)
        return {
            project.id: project.name
            + (
                f" (private, only visible to {project.author} and admins)"
                if not project.public and project.author != username
                else ""
            )
            for project in projects
        }
    else:
        public_projects = db.select(Project, public=True)
        private_projects = db.select(Project, public=False, author=username)
        extra_projects = db.select(Project, id__in=extra_project_ids or [])
        projects = sorted(public_projects + private_projects + extra_projects, key=lambda project: project.name.lower())
        return {project.id: project.name for project in projects}


@clear_when_modified(DescriptorValue)
@st.cache_data(max_entries=100, ttl="1h")
def get_cached_available_descriptors(design_ids: Collection[str]) -> dict[str, Descriptor]:
    """
    Return all descriptor keys found in DB for the given design ids.
    """
    return get_available_descriptors(design_ids)


@clear_when_modified(DescriptorValue)
@st.cache_data(max_entries=100, ttl="1h")
def get_cached_descriptor_values(descriptor_key: str, design_ids: list[str]) -> pd.Series():
    return db.select_descriptor_values(descriptor_key, design_ids=design_ids)


@clear_when_modified(Design)
@st.cache_data(max_entries=100, ttl="10m")
def get_cached_design(design_id: str) -> Design:
    return db.get(Design, id=design_id)


@clear_when_modified(Design)
@st.cache_data(max_entries=10, ttl="1h")
def get_cached_designs(design_ids: Collection[str]) -> list[Design]:
    designs_by_id = {d.id: d for d in db.select(Design, id__in=design_ids)}
    return [designs_by_id[design_id] for design_id in design_ids if design_id in designs_by_id]


@clear_when_modified(Design)
@st.cache_data(max_entries=10, ttl="1h")
def get_cached_design_ids(pool_ids: list[str], **filters) -> list[str]:
    """Get design ids matching the given filters, sorted from most recently created."""
    ids = db.select_values(Design, "id", pool_id__in=pool_ids, **filters)
    order = dict(zip(pool_ids, range(len(pool_ids))))
    return sorted(ids, key=lambda design_id: order.get(Design.design_id_to_pool_id(design_id)))


@clear_when_modified(Round)
@st.cache_data(max_entries=100, ttl="10m")
def get_cached_round(round_id: str) -> Round:
    return db.get(Round, id=round_id)


@clear_when_modified(Round)
@st.cache_data(max_entries=100, ttl="10m")
def get_cached_rounds(project_id: str = None) -> list[Round]:
    return db.select(Round, project_id=project_id, order_by="-created_date_utc")


@clear_when_modified(Pool)
@st.cache_data(max_entries=100, ttl="10m")
def get_cached_pool(pool_id: str) -> Pool:
    return db.get(Pool, id=pool_id)


@clear_when_modified(Pool)
@st.cache_data(max_entries=100, ttl="1h")
def get_cached_pools(pool_ids: Collection[str]) -> list[Pool]:
    return db.select(Pool, id__in=pool_ids, order_by="-created_date_utc")


@clear_when_modified(DesignJob)
@st.cache_data(max_entries=100, ttl="1h")
def get_cached_design_jobs(design_job_ids: Collection[str]) -> list[DesignJob]:
    return db.select(DesignJob, id__in=design_job_ids, order_by="-created_date_utc")


@clear_when_modified(DesignJob)
@st.cache_data(max_entries=100, ttl="10m")
def get_cached_design_job(design_job_id: str) -> DesignJob:
    return db.get(DesignJob, id=design_job_id)


def get_cached_design_jobs_table(round_ids: list[str]):
    design_job_ids = [v for v in db.select_values(Pool, "design_job_id", round_id__in=round_ids) if v]
    if db.count(DesignJob, id__in=design_job_ids, job_result=None):
        # if any jobs are still running, do not use cache so that job status is updated
        _get_cached_design_jobs_table.clear(round_ids=round_ids)
    return _get_cached_design_jobs_table(round_ids=round_ids)


@clear_when_modified(DesignJob)
@st.cache_data(show_spinner=False, max_entries=100, ttl="1m")  # update every minute since the table contains datetimes
def _get_cached_design_jobs_table(
    project_id: str = None, round_ids: list[str] = None, update=True, **pool_filters
) -> pd.DataFrame:
    return get_design_jobs_table(project_id=project_id, round_ids=round_ids, update=update, **pool_filters)


@clear_when_modified(Pool, Design)
@st.cache_data(max_entries=100, ttl="1m")  # update every minute since the table contains datetimes
def get_cached_pools_table(project_id: str = None, round_ids: list[str] = None) -> pd.DataFrame:
    return get_pools_table(project_id=project_id, round_ids=round_ids)


@clear_when_modified(DescriptorValue)
@st.cache_data(max_entries=100, ttl="1h")
def get_cached_design_descriptors(design_id: str, descriptor_keys: list[str]) -> pd.Series():
    return db.select_design_descriptors(design_id=design_id, descriptor_keys=descriptor_keys)


@clear_when_modified(Design)
@st.cache_data()
def get_cached_designs_accept_field(design_ids: list[str]) -> List[bool]:
    return db.get_design_accepted_values(design_ids)


@clear_when_modified(Pool, DesignJob)
@st.cache_data(ttl="1h")
def get_cached_workflow_pools_and_jobs(project_id: str, workflow_names: Collection[str]):
    round_ids = db.select_unique_values(Round, "id", project_id=project_id)
    pools = [pool for pool in db.select(Pool, round_id__in=round_ids) if pool.design_job_id]
    all_jobs_by_id = {
        j.id: j for j in db.select(DesignJob, id__in=[p.design_job_id for p in pools], order_by="-created_date_utc")
    }
    pools_by_id = {}
    jobs_by_id = {}
    for pool in pools:
        job = all_jobs_by_id[pool.design_job_id]
        if job.workflow.name not in workflow_names:
            continue
        pools_by_id[pool.id] = pool
        jobs_by_id[job.id] = job
    return pools_by_id, jobs_by_id



import streamlit as st


def is_test_dialog_shown(dialog_key: str):
    return dialog_key in st.session_state.get("test_dialogs", [])


def show_test_dialog(at, dialog_key: str):
    if "test_dialogs" not in at.session_state:
        at.session_state["test_dialogs"] = []
    at.session_state["test_dialogs"].append(dialog_key)
    at.run()


def hide_test_dialog(at, dialog_key: str):
    if "test_dialogs" not in at.session_state:
        at.session_state["test_dialogs"] = []
    at.session_state["test_dialogs"].remove(dialog_key)
    at.run()



from ovo.core.database.base_db import DBEngine
from ovo.core.database.sql_db import SqlDBEngine
from ovo.core.database.models import *  # limited with __all__ in models.py
from ovo.core.database import models_bindcraft
from ovo.core.database import models_rfdiffusion
from ovo.core.database import models_proteinqc
from ovo.core.database import models_refolding
from ovo.core.database import descriptors
from ovo.core.database import descriptors_refolding
from ovo.core.database import descriptors_proteinqc
from ovo.core.database import descriptors_rfdiffusion
from ovo.core.database import descriptors_bindcraft



import os
import re
from datetime import datetime
from ovo.core.auth import get_username
from ovo.core.utils.formatting import format_duration
from ovo.core.scheduler.base_scheduler import Scheduler, SchedulerTypes
from ovo.core.aws import AWSSessionManager


@SchedulerTypes.register()
class HealthOmicsScheduler(Scheduler):
    def __init__(
        self,
        name: str,
        workdir: str,
        reference_files_dir: str,
        aws: AWSSessionManager,
        allow_submit: bool = True,
        submission_args: dict = None,
    ):
        """
        Args:
            aws:
            name: Human-readable label for this scheduler
            workdir: Working directory - where to store workflow outputs (S3 URI including s3:// prefix)
            reference_files_dir: UNUSED Directory with model weights and other reference files
            role_arn: AWS IAM role ARN
            allow_submit: Whether to allow job submission
            submission_args: Default submission arguments, can be overridden in submit method
        """
        super().__init__(name, workdir, reference_files_dir, allow_submit, submission_args)
        self.aws = aws

    def submit(self, pipeline_name: str, params: dict = None, submission_args: dict = None) -> str:
        """
        Submits a workflow asynchronously and returns the PID.

        :param pipeline_name: Workflow name and revision (ovo.rfdiffusion-end-to-end or github url with @version)
        :param params: Dictionary of parameters to pass to the workflow.
        :param submission_args: Submission arguments for the scheduler, overrides values in self.submission_args
        :return: Scheduler job ID
        """
        assert isinstance(params, dict), f"params should be a dictionary, got: {type(params).__name__}"

        if not self.allow_submit:
            raise RuntimeError("Job submission is disabled")

        submission_args = {**self.submission_args, **(submission_args or {})}
        workflow_name_prefix = submission_args.pop("workflow_name_prefix", "")
        role_arn = submission_args.pop("role_arn", "")

        if re.fullmatch(r"https?://.*", pipeline_name):
            raise NotImplementedError("GitHub URL pipeline names are not supported yet with HealthOmicsScheduler")
        elif "/" in pipeline_name or pipeline_name == "." or pipeline_name.endswith(".nf"):
            raise NotImplementedError("Local path pipeline names are not supported with HealthOmicsScheduler")
        elif "." in pipeline_name:  # custom workflow path (python_module_name.my_workflow)
            module_name, workflow_subpath = pipeline_name.split(".")
        else:
            module_name = "ovo"
            workflow_subpath = pipeline_name

        omics_workflow_name = workflow_name_prefix + module_name + "_" + workflow_subpath
        workflow_id = self.aws.get_latest_workflow_id(workflow_name=omics_workflow_name)

        if submission_args:
            # submission args should be empty at this point
            raise ValueError(f"Unrecognized {type(self).__name__} submission args: {submission_args.keys()}")

        # AWS params
        params = params.copy()
        params["account_id"] = self.aws.get_account_id()
        params["aws_region"] = self.aws.region_name
        params["workflow_bucket"] = self.workdir.replace("s3://", "").split("/")[0]
        params["reference_files_dir"] = "unused"

        username = get_username()

        # Submit the job
        response = self.aws.omics.start_run(
            roleArn=role_arn,
            workflowId=workflow_id,
            workflowType="PRIVATE",
            name=f"OVO {omics_workflow_name} {username}",
            outputUri=f"{self.workdir.rstrip('/')}/{workflow_subpath}/",
            parameters=params,
            storageCapacity=500,
            logLevel="ALL",
            tags={"user": username, "app": "OVO"},
            storageType="STATIC",
        )

        return response["id"]

    def get_status_label(self, job_id: str) -> str:
        """Get human-readable job status label. Should NOT be used to determine job status.

        Can return "Cancelled", "Completed", "Deleted", "Failed", "Pending", "Running", "Starting", "Stopping"
        """
        return self.aws.omics.get_run(id=job_id)["status"].capitalize()

    def get_result(self, job_id: str) -> str | bool | None:
        """Get job result: True if successful, False if failed, None if running."""
        if not isinstance(job_id, str):
            raise ValueError(
                f"Job ID of {type(self).__name__} should be a string, got {job_id} ({type(job_id).__name__}). Make sure to use job_id instead of the DB id."
            )
        status = self.aws.omics.get_run(id=job_id)["status"]
        if status in ["STARTING", "PENDING", "PENDING_REDRIVE", "RUNNING", "STOPPING"]:
            return None
        if status in ["SUCCEEDED", "COMPLETED"]:
            return True
        return False

    def get_log(self, job_id: str) -> str | None:
        """Get job execution full log"""
        if not isinstance(job_id, str):
            raise ValueError(
                f"Job ID of {type(self).__name__} should be a string, got {job_id} ({type(job_id).__name__}). Make sure to use job_id instead of the DB id."
            )
        run = self.aws.omics.get_run(id=job_id)
        # TODO this log file is only available when job is finished
        # We need to read this from CloudWatch instead
        # See logLocation: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/omics/client/get_run.html
        engine_log_s3_uri = os.path.join(run["runOutputUri"], "logs", "engine.log")
        if self.aws.s3_file_exists(engine_log_s3_uri):
            return self.aws.get_s3_bytes(engine_log_s3_uri).decode("utf-8")
        return None

    def cancel(self, job_id):
        """Cancel job execution"""
        raise NotImplementedError()

    def get_output_dir(self, job_id: str) -> str:
        """Get job output path"""
        return os.path.join(self.aws.omics.get_run(id=job_id)["runOutputUri"], "pubdir")

    def get_job_start_time(self, job_id: str) -> datetime | None:
        """Get job start time"""
        run = self.aws.omics.get_run(id=job_id)
        if "startTime" in run:
            return run["startTime"].replace(tzinfo=None)
        return None

    def get_job_stop_time(self, job_id: str) -> datetime | None:
        """Get job end time"""
        run = self.aws.omics.get_run(id=job_id)
        if "stopTime" in run:
            return run["stopTime"].replace(tzinfo=None)
        return None

    def get_startup_time_minutes(self):
        """Get startup time of a task (in minutes)"""
        return 10



import streamlit as st
import os
from molstar_custom_component import molstar_custom_component
from molstar_custom_component.contigs_parser import ContigsParser
from molstar_custom_component.dataclasses import StructureVisualization, ChainVisualization

# `$ python -m streamlit run molstar_custom_component/main.py`
# do not forget to have venv activated

# pdb_path = st.text_input("Enter a PDB URL/filepath", value="http://localhost:5500/ovo_molstar/test_files/multichain_scaffold/test_0.pdb")
# trb_path = st.text_input("Enter a .trb URL/filepath", value="http://localhost:5500/ovo_molstar/test_files/multichain_scaffold/test_0.trb")

# contigs = st.text_input("Specify the contigs (not used when .trb file provided)", value="A117-125/10-20/A107-114/10-20/A47-53/10-20/A60-65")

paths = [
    "http://localhost:5500/ovo/app/components/molstar_custom_component/test_files/multichain_more/result_0",
    "http://localhost:5500/ovo/app/components/molstar_custom_component/test_files/multichain_binder/result_0",
    "http://localhost:5500/ovo/app/components/molstar_custom_component/test_files/multichain_scaffold/test_0",
]

for path in paths:
    pdb_path = path + ".pdb"
    trb_path = path + ".trb"

    print(pdb_path)

    if os.path.exists(pdb_path):
        pdb = open(pdb_path).read()
    else:
        pdb = pdb_path

    parser = ContigsParser()
    contigs = parser.parse_contigs_trb(trb_path)

    cp = molstar_custom_component(
        structures=[
            StructureVisualization(
                pdb=pdb,
                highlighted_selections=["A22-25", "A28-30"],
                color="hydrophobicity",
                representation_type="gaussian-surface",
                contigs=contigs,
            ),
            StructureVisualization(
                pdb="http://localhost:5500/ovo/app/components/molstar_custom_component/test_files/multichain_scaffold/test_0.pdb",
                highlighted_selections=["A22-25", "A28-30"],
                chains=[ChainVisualization(chain_id="A", color="uniform", color_params={"value": "0x00ff00"})],
            ),
        ],
        selection_mode=True,
        show_controls=True,
        key="example",
    )

    st.write(cp)
